import asyncio
from navconfig import BASE_DIR
from parrot.llms.vertex import VertexLLM
from parrot.loaders import (
    PDFLoader
)

async def process_pdf():
    llm = VertexLLM(
        model='gemini-1.5-flash-001',
        temperature=0.1,
        top_k=30,
        Top_p=0.5,
    )
    # Add LLM
    doc = BASE_DIR.joinpath('documents', 'AR_Certification_Skill_Practice_Scorecard_EXAMPLE.pdf')
    print(':: Processing: ', doc)
    # PDF Files
    loader = PDFLoader(
        doc,
        source_type=f"PDF {doc.name}",
        llm=llm.get_llm(),
        language="en",
        parse_images=False,
        page_as_images=True
    )
    docs = loader.load()
    print('DOCS > ', docs)

if __name__ == "__main__":
    agent = asyncio.run(process_pdf())



from collections.abc import Callable
from typing import Any
from abc import abstractmethod
from pathlib import Path, PurePath
from PIL import Image
from .abstract import AbstractLoader
from ..conf import STATIC_DIR


class BasePDF(AbstractLoader):
    """
    Base Abstract loader for all PDF files.
    """
    _extension = ['.pdf']

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type=source_type, **kwargs)
        self.path = path
        if isinstance(path, str):
            self.path = Path(path).resolve()
        self.save_images: bool = bool(kwargs.get('save_images', False))
        self._imgdir = STATIC_DIR.joinpath('images')
        if self.save_images is True:
            if self._imgdir.exists() is False:
                self._imgdir.mkdir(parents=True, exist_ok=True)
        if language == 'en':
            language = 'eng'
        self._lang = language

    def save_image(self, img_stream: Image, image_name: str, save_path: Path):
        # Create the image directory if it does not exist
        if save_path.exists() is False:
            save_path.mkdir(parents=True, exist_ok=True)
        img_path = save_path.joinpath(image_name)
        self.logger.notice(
            f"Saving Image Page on {img_path}"
        )
        if not img_path.exists():
            # Save the image
            img_stream.save(img_path, format="PNG", optimize=True)
        return img_path

    @abstractmethod
    def _load_pdf(self, path: Path) -> list:
        """
        Load a PDF file using Fitz.

        Args:
            path (Path): The path to the PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        pass

    def load(self) -> list:
        """
        Load data from a PDF file.

        Args:
            source (str): The path to the PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        if isinstance(self.path, list):
            # list of files:
            documents = []
            for p in self.path:
                if self._check_path(p):
                    documents.extend(self._load_pdf(p))
        if not self.path.exists():
            raise FileNotFoundError(
                f"PDF file/directory not found: {self.path}"
            )
        if self.path.is_dir():
            documents = []
            # iterate over the files in the directory
            for ext in self._extension:
                for item in self.path.glob(f'*{ext}'):
                    if set(item.parts).isdisjoint(self.skip_directories):
                        documents.extend(self._load_pdf(item))
        elif self.path.is_file():
            documents = self._load_pdf(self.path)
        else:
            raise ValueError(
                f"PDF Loader: Invalid path: {self.path}"
            )
        return self.split_documents(documents)

    def parse(self, source):
        raise NotImplementedError(
            "Parser method is not implemented for PDFLoader."
        )



# cython: language_level=3, embedsignature=True, boundscheck=False, wraparound=True, initializedcheck=False
# Copyright (C) 2018-present Jesus Lara
#
cdef class ParrotError(Exception):
    """Base class for other exceptions"""

    def __init__(self, object message, *args, **kwargs):
        super().__init__(message)
        if hasattr(message, 'message'):
            self.message = message.message
        else:
            self.message = str(message)
        self.stacktrace = None
        if 'stacktrace' in kwargs:
            self.stacktrace = kwargs['stacktrace']
        self.args = kwargs

    def __repr__(self):
        return f"{self.message}"

    def __str__(self):
        return f"{self.message}"

    def get(self):
        return self.message

cdef class ConfigError(ParrotError):
    pass



from typing import Optional, Union
from pytube import YouTube
from youtube_transcript_api import NoTranscriptFound
import torch
from langchain.docstore.document import Document
from langchain_community.document_loaders.parsers.audio import (
    OpenAIWhisperParserLocal
)
from langchain_community.document_loaders import YoutubeLoader as YTLoader
from langchain_community.document_loaders.generic import (
    GenericLoader
)
from langchain_community.document_loaders.blob_loaders.youtube_audio import (
    YoutubeAudioLoader
)
from .video import VideoLoader


def extract_video_id(url):
    parts = url.split("?v=")
    video_id = parts[1].split("&")[0]
    return video_id


class YoutubeLoader(VideoLoader):
    """
    Loader for Youtube videos.
    """
    _extension = ['.youtube']
    encoding = 'utf-8'
    chunk_size = 768

    def get_video_info(self, url: str) -> dict:
        yt = YouTube(url)
        return  {
            "url": url,
            "video_id": yt.video_id,
            "watch_url": yt.watch_url,
            "embed_url": yt.embed_url,
            "title": yt.title or "Unknown",
            "description": yt.description or "Unknown",
            "view_count": yt.views or 0,
            #"thumbnail_url": yt.thumbnail_url or "Unknown",
            "publish_date": yt.publish_date.strftime("%Y-%m-%d %H:%M:%S") if yt.publish_date else "Unknown",
            # "length": yt.length or 0,
            "author": yt.author or "Unknown",
        }

    def load_video(self, url: str, video_title: str, transcript: Optional[Union[str, None]] = None) -> list:
        # first: load video metadata:
        video_info = self.get_video_info(url)
        # Second: load video transcript (if any)
        if transcript is None:
            try:
                documents = []
                docs = []
                # first: download video
                file_path = self.download_video(url, self._video_path)
                audio_path = file_path.with_suffix('.mp3')
                transcript_path = file_path.with_suffix('.vtt')
                # second: extract audio
                self.extract_audio(file_path, audio_path)
                transcript_whisper = self.get_whisper_transcript(audio_path)
                transcript = transcript_whisper['text']
                # Summarize the transcript
                try:
                    summary = self.get_summary_from_text(transcript)
                except Exception:
                    summary = ''
                # Create Two Documents, one is for transcript, second is VTT:
                metadata = {
                    "url": f"{url}",
                    "source": f"{url}",
                    "index": video_title,
                    "filename": video_title,
                    "question": '',
                    "answer": '',
                    "source_type": self._source_type,
                    'type': 'video_transcript',
                    "summary": f"{summary!s}",
                    "document_meta": {
                        "language": self._language,
                        "title": video_title,
                        "docinfo": video_info
                    }
                }
                if self.topics:
                    metadata["document_meta"]['topic_tags'] = self.topics
                doc = Document(
                    page_content=transcript,
                    metadata=metadata
                )
                documents.append(doc)
                # VTT version:
                transcript = self.transcript_to_vtt(transcript_whisper, transcript_path)
                if transcript:
                    doc = Document(
                        page_content=transcript,
                        metadata=metadata
                    )
                    documents.append(doc)
                # Saving every dialog chunk as a separate document
                dialogs = self.transcript_to_blocks(transcript_whisper)
                for chunk in dialogs:
                    _meta = {
                        "index": f"{video_title}:{chunk['id']}",
                        "document_meta": {
                            "start": f"{chunk['start_time']}",
                            "end": f"{chunk['end_time']}",
                            "id": f"{chunk['id']}",
                            "language": self._language,
                            "title": video_title,
                            "topic_tags": ""
                        }
                    }
                    _info = {**metadata, **_meta}
                    doc = Document(
                        page_content=chunk['text'],
                        metadata=_info
                    )
                    docs.append(doc)
                documents.extend(docs)
                return documents
            except Exception:
                try:
                    loader = YTLoader.from_youtube_url(
                        url,
                        add_video_info=True,
                        language=[self._language],
                    )
                    docs = loader.load()
                except NoTranscriptFound:
                    loader = GenericLoader(
                        YoutubeAudioLoader([url], str(self._video_path)),
                        OpenAIWhisperParserLocal(
                            # lang_model='openai/whisper-medium.en',
                            lang_model='openai/whisper-small.en',
                            device=self._get_device()
                        )
                    )
                    docs = loader.load()
                if not docs:
                    self.logger.warning(
                        f"Unable to load Youtube Video {url}"
                    )
                    return []
                summary = self.get_summary(
                    docs
                )
                for doc in docs:
                    doc.metadata['source_type'] = self._source_type
                    doc.metadata['summary'] = f"{summary!s}"
                    doc.metadata['index'] = ''
                    doc.metadata['filename'] = ''
                    doc.metadata['question'] = ''
                    doc.metadata['answer'] = ''
                    doc.metadata['type'] = 'video_transcript'
                    doc.metadata['document_meta'] = {}
                    if self.topics:
                        doc.metadata['document_meta']['topic_tags'] = self.topics
                    # add video metadata to document metadata:
                    for key, value in video_info.items():
                        doc.metadata['document_meta'][key] = f"{value!s}"
                return docs
        else:
            with open(transcript, 'r') as f:
                transcript = f.read()
            if transcript:
                summary = self.get_summary_from_text(transcript)
                transcript_whisper = None
                metadata = {
                    "source": url,
                    "url": url,
                    "index": '',
                    "filename": '',
                    "question": '',
                    "answer": '',
                    "source_type": self._source_type,
                    'type': 'video_transcript',
                    'summary': f"{summary!s}",
                    "document_meta": {
                        "language": self._language,
                        "title": video_title
                    },
                }
                if self.topics:
                    metadata['document_meta']['topic_tags'] = self.topics
                doc = Document(
                    page_content=transcript,
                    metadata=metadata
                )
                return [doc]



from typing import Optional, Union, Any
import asyncio
import uuid
import torch
from pymilvus import (
    MilvusClient,
    Collection,
    FieldSchema,
    CollectionSchema,
    DataType,
    connections,
    db
)
from pymilvus.exceptions import MilvusException
from langchain_milvus import Milvus  # pylint: disable=import-error, E0611
from langchain.memory import VectorStoreRetrieverMemory
from .abstract import AbstractStore
from ..conf import (
    MILVUS_HOST,
    MILVUS_PROTOCOL,
    MILVUS_PORT,
    MILVUS_URL,
    MILVUS_TOKEN,
    MILVUS_USER,
    MILVUS_PASSWORD,
    MILVUS_SECURE,
    MILVUS_SERVER_NAME,
    MILVUS_CA_CERT,
    MILVUS_SERVER_CERT,
    MILVUS_SERVER_KEY,
    MILVUS_USE_TLSv2
)


class MilvusConnection:
    """
    Context Manager for Milvus Connections.
    """
    def __init__(self, alias: str = 'default', **kwargs):
        self._connected: bool = False
        self.kwargs = kwargs
        self.alias: str = alias

    def connect(self, alias: str = None, **kwargs):
        if not alias:
            alias = self.alias
        conn = connections.connect(
            alias=alias,
            **kwargs
        )
        self._connected = True
        return alias

    def is_connected(self):
        return self._connected

    def close(self, alias: str = None):
        try:
            connections.disconnect(alias=alias)
        finally:
            self._connected = False

    def __enter__(self):
        self.connect(alias=self.alias, **self.kwargs)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close(alias=self.alias)
        return self


class MilvusStore(AbstractStore):
    """MilvusStore class.

    Milvus is a Vector Database multi-layered.

    Args:
        host (str): Milvus host.
        port (int): Milvus port.
        url (str): Milvus URL.
    """

    def __init__(self, embeddings = None, **kwargs):
        super().__init__(embeddings, **kwargs)
        self.use_bge: bool = kwargs.pop("use_bge", False)
        self.fastembed: bool = kwargs.pop("use_fastembed", False)
        self.database: str = kwargs.pop('database', '')
        self.collection = kwargs.pop('collection_name', '')
        self.dimension: int = kwargs.pop("dimension", 768)
        self._metric_type: str = kwargs.pop("metric_type", 'COSINE')
        self._index_type: str = kwargs.pop("index_type", 'IVF_FLAT')
        self.host = kwargs.pop("host", MILVUS_HOST)
        self.port = kwargs.pop("port", MILVUS_PORT)
        self.protocol = kwargs.pop("protocol", MILVUS_PROTOCOL)
        self.create_database: bool = kwargs.pop('create_database', True)
        self.url = kwargs.pop("url", MILVUS_URL)
        self._client_id = kwargs.pop('client_id', 'default')
        if not self.url:
            self.url = f"{self.protocol}://{self.host}:{self.port}"
        else:
            # Extract host and port from URL
            if not self.host:
                self.host = self.url.split("://")[-1].split(":")[0]
            if not self.port:
                self.port = int(self.url.split(":")[-1])
        self.token = kwargs.pop("token", MILVUS_TOKEN)
        # user and password (if required)
        self.user = kwargs.pop("user", MILVUS_USER)
        self.password = kwargs.pop("password", MILVUS_PASSWORD)
        # SSL/TLS
        self._secure: bool = kwargs.pop('secure', MILVUS_SECURE)
        self._server_name: str = kwargs.pop('server_name', MILVUS_SERVER_NAME)
        self._cert: str = kwargs.pop('server_pem_path', MILVUS_SERVER_CERT)
        self._ca_cert: str = kwargs.pop('ca_pem_path', MILVUS_CA_CERT)
        self._cert_key: str = kwargs.pop('client_key_path', MILVUS_SERVER_KEY)
        # Any other argument will be passed to the Milvus client
        self.kwargs = {
            "uri": self.url,
            "host": self.host,
            "port": self.port,
            **kwargs
        }
        if self.token:
            self.kwargs['token'] = self.token
        if self.user:
            self.kwargs['token'] = f"{self.user}:{self.password}"
        # SSL Security:
        if self._secure is True:
            args = {
                "secure": self._secure,
                "server_name": self._server_name
            }
            if self._cert:
                if MILVUS_USE_TLSv2 is True:
                    args['client_pem_path'] = self._cert
                    args['client_key_path'] = self._cert_key
                else:
                    args["server_pem_path"] = self._cert
            if self._ca_cert:
                args['ca_pem_path'] = self._ca_cert
            self.kwargs = {**self.kwargs, **args}
        # 1. Check if database exists:
        if self.database:
            self.kwargs['db_name'] = self.database
            self.use_database(
                self.database,
                create=self.create_database
            )

    async def __aenter__(self):
        try:
            self.tensor = torch.randn(1000, 1000).cuda()
        except RuntimeError:
            self.tensor = None
        if self._embed_ is None:
            self._embed_ = self.create_embedding(
                model_name=self.embedding_name
            )
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        # closing Embedding
        self._embed_ = None
        del self.tensor
        try:
            self.close(alias=self._client_id)
            torch.cuda.empty_cache()
        except RuntimeError:
            pass

    def connection(self, alias: str = None):
        if not alias:
            # self._client_id = str(uuid.uuid4())
            self._client_id = 'uri-connection'
        else:
            self._client_id = alias
        # making the connection:
        self._client, self._client_id = self.connect(
            alias=self._client_id
        )
        return self

    def connect(self, alias: str = None) -> tuple:
        # 1. Set up a pyMilvus default connection
        # Unique connection:
        if not alias:
            alias = "default"
        _ = connections.connect(
            alias=alias,
            **self.kwargs
        )
        client = MilvusClient(
            **self.kwargs
        )
        self._connected = True
        return client, alias

    def close(self, alias: str = "default"):
        connections.disconnect(alias=alias)
        try:
            self._client.close()
        finally:
            self._connected = False

    def create_db(self, db_name: str, alias: str = 'default', **kwargs) -> bool:
        args = {
            "uri": self.url,
            "host": self.host,
            "port": self.port,
            **kwargs
        }
        try:
            conn = connections.connect(alias, **args)
            db.create_database(db_name)
            self.logger.notice(
                f"Database {db_name} created successfully."
            )
        except Exception as e:
            raise ValueError(
                f"Error creating database: {e}"
            )
        finally:
            connections.disconnect(alias="uri-connection")

    def use_database(
        self,
        db_name: str,
        alias:str = 'default',
        create: bool = False
    ) -> None:
        try:
            conn = connections.connect(alias, **self.kwargs)
        except MilvusException as exc:
            if "database not found" in exc.message:
                args = self.kwargs.copy()
                del args['db_name']
                self.create_db(db_name, alias=alias, **args)
        # re-connect:
        try:
            _ = connections.connect(alias, **self.kwargs)
            if db_name not in db.list_database(using=alias):
                if self.create_database is True or create is True:
                    try:
                        db.create_database(db_name, using=alias, timeout=10)
                        self.logger.notice(
                            f"Database {db_name} created successfully."
                        )
                    except Exception as e:
                        raise ValueError(
                            f"Error creating database: {e}"
                        )
                else:
                    raise ValueError(
                        f"Database {db_name} does not exist."
                    )
        finally:
            connections.disconnect(alias=alias)

    def setup_vector(self):
        self.vector = Milvus(
            self._embed_,
            consistency_level='Bounded',
            connection_args={**self.kwargs},
            collection_name=self.collection,
        )
        return self.vector

    def get_vectorstore(self):
        return self.get_vector()

    def collection_exists(self, collection_name: str) -> bool:
        if collection_name in self._client.list_collections():
            return True
        return False

    def check_state(self, collection_name: str) -> dict:
        return self._client.get_load_state(collection_name=collection_name)

    async def delete_collection(self, collection: str = None) -> dict:
        self._client.drop_collection(
            collection_name=collection
        )

    async def create_collection(
        self,
        collection_name: str,
        document: Any = None,
        dimension: int = 768,
        index_type: str = None,
        metric_type: str = None,
        schema_type: str = 'default',
        metadata_field: str = None,
        **kwargs
    ) -> dict:
        """create_collection.

        Create a Schema (Milvus Collection) on the Current Database.

        Args:
            collection_name (str): Collection Name.
            document (Any): List of Documents.
            dimension (int, optional): Vector Dimension. Defaults to 768.
            index_type (str, optional): Default index type of Vector Field. Defaults to "HNSW".
            metric_type (str, optional): Default Metric for Vector Index. Defaults to "L2".
            schema_type (str, optional): Description of Model. Defaults to 'default'.

        Returns:
            dict: _description_
        """
        # Check if collection exists:
        if self.collection_exists(collection_name):
            self.logger.warning(
                f"Collection {collection_name} already exists."
            )
            return None
        idx_params = {}
        if not index_type:
            index_type = self._index_type
        if index_type == 'HNSW':
            idx_params = {
                "M": 36,
                "efConstruction": 1024
            }
        elif index_type in ('IVF_FLAT', 'SCANN', 'IVF_SQ8'):
            idx_params = {
                "nlist": 1024
            }
        elif index_type in ('IVF_PQ'):
            idx_params = {
                "nlist": 1024,
                "m": 16
            }
        if not metric_type:
            metric_type = self._metric_type  # default metric type
        if schema_type == 'default':
            # Default Collection for all loaders:
            schema = MilvusClient.create_schema(
                auto_id=False,
                enable_dynamic_field=True,
                description=collection_name
            )
            schema.add_field(
                field_name="pk",
                datatype=DataType.INT64,
                is_primary=True,
                auto_id=True,
                max_length=100
            )
            schema.add_field(
                field_name="index",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="url",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="source",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="filename",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="question",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="answer",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="source_type",
                datatype=DataType.VARCHAR,
                max_length=128
            )
            schema.add_field(
                field_name="type",
                datatype=DataType.VARCHAR,
                max_length=65535
            )
            schema.add_field(
                field_name="text",
                datatype=DataType.VARCHAR,
                description="Text",
                max_length=65535
            )
            schema.add_field(
                field_name="summary",
                datatype=DataType.VARCHAR,
                description="Summary (refine resume)",
                max_length=65535
            )
            schema.add_field(
                field_name="vector",
                datatype=DataType.FLOAT_VECTOR,
                dim=dimension,
                description="vector"
            )
            # schema.add_field(
            #     field_name="embedding",
            #     datatype=DataType.FLOAT_VECTOR,
            #     dim=dimension,
            #     description="Binary Embeddings"
            # )
            schema.add_field(
                field_name="document_meta",
                datatype=DataType.JSON,
                description="Custom Metadata information"
            )
            index_params = self._client.prepare_index_params()
            index_params.add_index(
                field_name="pk",
                index_type="STL_SORT"
            )
            index_params.add_index(
                field_name="text",
                index_type="marisa-trie"
            )
            index_params.add_index(
                field_name="summary",
                index_type="marisa-trie"
            )
            index_params.add_index(
                field_name="vector",
                index_type=index_type,
                metric_type=metric_type,
                params=idx_params
            )
            self._client.create_collection(
                collection_name=collection_name,
                schema=schema,
                index_params=index_params,
                num_shards=2
            )
            await asyncio.sleep(2)
            res = self._client.get_load_state(
                collection_name=collection_name
            )
            return None
        else:
            self._client.create_collection(
                collection_name=collection_name,
                dimension=dimension
            )
            if metadata_field:
                kwargs['metadata_field'] = metadata_field
            # Here using drop_old=True to force recreate based on the first document
            docstore = Milvus.from_documents(
                [document],  # Only the first document
                self._embed_,
                connection_args={**self.kwargs},
                collection_name=collection_name,
                drop_old=True,
                # consistency_level='Session',
                primary_field='pk',
                text_field='text',
                vector_field='vector',
                **kwargs
            )
            return docstore

    async def load_documents(
        self,
        documents: list,
        collection: str = None,
        upsert: bool = False,
        attribute: str = 'source_type',
        metadata_field: str = None,
        **kwargs
    ):
        if not collection:
            collection = self.collection
        try:
            tensor = torch.randn(1000, 1000).cuda()
        except Exception:
            tensor = None
        if upsert is True:
            # get first document
            doc = documents[0]
            # getting source type:
            doc_type = doc.metadata.get('attribute', None)
            if attribute:
                deleted = self._client.delete(
                    collection_name=collection,
                    filter=f'{attribute} == "{doc_type}"'
                )
                self.logger.notice(
                    f"Deleted documents with {attribute} {attribute}: {deleted}"
                )
        if metadata_field:
            # document_meta
            kwargs['metadata_field'] = metadata_field
        docstore = Milvus.from_documents(
            documents,
            self._embed_,
            connection_args={**self.kwargs},
            collection_name=collection,
            consistency_level='Bounded',
            drop_old=False,
            primary_field='pk',
            text_field='text',
            vector_field='vector',
            **kwargs
        )
        del tensor
        return docstore

    def upsert(self, payload: dict, collection: str = None) -> None:
        pass

    def insert(
        self,
        payload: Union[dict, list],
        collection: Union[str, None] = None
    ) -> dict:
        if collection is None:
            collection = self.collection
        result = self._client.insert(
            collection_name=collection,
            data=payload
        )
        collection.flush()
        return result

    def get_vector(
        self,
        collection: Union[str, None] = None,
        metric_type: str = None,
        nprobe: int = 32,
        metadata_field: str = None,
        consistency_level: str = 'session'
    ) -> Milvus:
        if not metric_type:
            metric_type = self._metric_type
        if not collection:
            collection = self.collection
        _search = {
            "search_params": {
                "metric_type": metric_type,
                "params": {"nprobe": nprobe, "nlist": 1024},
            }
        }
        if metadata_field:
            # document_meta
            _search['metadata_field'] = metadata_field
        _embed_ = self.create_embedding(
            model_name=self.embedding_name
        )
        return Milvus(
            embedding_function=_embed_,
            collection_name=collection,
            consistency_level=consistency_level,
            connection_args={
                **self.kwargs
            },
            primary_field='pk',
            text_field='text',
            vector_field='vector',
            **_search
        )

    def similarity_search(
        self,
        query: str,
        collection: Union[str, None] = None,
        limit: int = 2,
        consistency_level: str = 'Bounded'
    ) -> list:
        if collection is None:
            collection = self.collection
        if self._embed_ is None:
            _embed_ = self.create_embedding(
                model_name=self.embedding_name
            )
        else:
            _embed_ = self._embed_
        vector_db = Milvus(
            embedding_function=_embed_,
            collection_name=collection,
            consistency_level=consistency_level,
            connection_args={
                **self.kwargs
            },
            primary_field='pk',
            text_field='text',
            vector_field='vector'
        )
        return vector_db.similarity_search(query, k=limit)

    def search(
        self,
        payload: Union[dict, list],
        collection: Union[str, None] = None,
        limit: Optional[int] = None
    ) -> list:
        args = {}
        if collection is None:
            collection = self.collection
        if limit is not None:
            args = {"limit": limit}
        if isinstance(payload, dict):
            payload = [payload]
        result = self._client.search(
            collection_name=collection,
            data=payload,
            **args
        )
        return result

    def memory_retriever(self, num_results: int  = 5) -> VectorStoreRetrieverMemory:
        vectordb = Milvus.from_documents(
            {},
            self._embed_,
            connection_args={**self.kwargs}
        )
        retriever = Milvus.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)

    def save_context(self, memory: VectorStoreRetrieverMemory, context: list) -> None:
        for val in context:
            memory.save_context(val)



import os
from navconfig import config, BASE_DIR
from google.cloud import aiplatform
from langchain_google_vertexai import (
    ChatVertexAI,
    VertexAI,
    VertexAIModelGarden,
    VertexAIEmbeddings
)
from .abstract import AbstractLLM

class VertexLLM(AbstractLLM):
    """VertexLLM.

    Interact with VertexAI Language Model.

    Returns:
        _type_: VertexAI LLM.
    """
    model: str = "gemini-1.0-pro"
    embed_model: str = "textembedding-gecko@003"
    max_tokens: int = 1024
    supported_models: list = [
        "gemini-1.0-pro",
        "gemini-1.5-pro-001",
        "gemini-1.5-pro-exp-0801",
        "gemini-1.5-flash-preview-0514",
        "gemini-1.5-flash-001",
        "chat-bison@001",
        "claude-3-opus@20240229",
        'claude-3-5-sonnet@20240620'
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        use_garden: bool = kwargs.get("use_garden", False)
        project_id = config.get("VERTEX_PROJECT_ID")
        region = config.get("VERTEX_REGION")
        config_file = config.get('GOOGLE_CREDENTIALS_FILE', 'env/google/vertexai.json')
        config_dir = BASE_DIR.joinpath(config_file)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = str(config_dir)
        self.args = {
            "project": project_id,
            "location": region,
            "max_output_tokens": self.max_tokens,
            "temperature": self.temperature,
            "max_retries": 4,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "verbose": True,
        }
        if use_garden is True:
            base_llm = VertexAIModelGarden
            self.args['endpoint_id'] = self.model
        elif self.model == "chat":
            self.args['model_name'] = "chat-bison@001"
            base_llm = ChatVertexAI
        else:
            self.args['model_name'] = self.model
            base_llm = VertexAI
        # LLM
        self._llm = base_llm(
            system_prompt="Always respond in the same language as the user's question. If the user's language is not English, translate your response into their language.",
            **self.args
        )
        # Embedding Model:
        embed_model = kwargs.get("embed_model", self.embed_model)
        self._embed = VertexAIEmbeddings(
            model_name=embed_model,
            project=project_id,
            location=region,
            request_parallelism=5,
            max_retries=4,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )
        self._version_ = aiplatform.__version__



import asyncio
from parrot.chatbots.basic import Chatbot
from parrot.llms.groq import GroqLLM
from parrot.llms.vertex import VertexLLM

async def get_agent():
    """Return the New Agent.
    """
    # llm = GroqLLM(
    #     model="llama3-groq-70b-8192-tool-use-preview",
    #     temperature=0.1,
    #     top_k=30,
    #     Top_p=0.6,
    # )
    llm = VertexLLM(
        model='gemini-1.5-flash-001',
        temperature=0.1,
        top_k=30,
        Top_p=0.5,
    )
    agent = Chatbot(
        name='RFPBot',
        llm=llm
    )
    await agent.configure()
    return agent


async def ask_agent(conversation, query):
    return await conversation.invoke(query)

if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    agent = loop.run_until_complete(get_agent())
    query = input("Type in your query: \n")
    EXIT_WORDS = ["exit", "quit", "bye"]
    memory = agent.get_memory(key='chat_history')
    while query not in EXIT_WORDS:
        if query:
            with agent.get_retrieval() as retrieval:
                conversation = retrieval.conversation(
                    question=query,
                    search_kwargs={"k": 10},
                    memory=memory
                )
                response = loop.run_until_complete(
                    ask_agent(conversation, query)
                )
                print('::: Response: ', response.response)

        query = input("Type in your query: \n")



from langchain_community.llms import HuggingFacePipeline  # pylint: disable=import-error, E0611
from transformers import AutoModelForCausalLM, AutoTokenizer
from .abstract import AbstractLLM

class HuggingFace(AbstractLLM):
    """HuggingFace.

    Load a LLM (Language Model) from HuggingFace Hub.

    Only supports text-generation, text2text-generation, summarization and translation for now.

    Returns:
        _type_: an instance of HuggingFace LLM Model.
    """
    model: str = "databricks/dolly-v2-3b"
    embed_model: str = None
    max_tokens: int = 1024
    supported_models: list = [
        "databricks/dolly-v2-3b",
        "gpt2",
        "bigscience/bloom-1b7",
        "meta-llama/Llama-2-7b-hf"
    ]

    def __init__(self, *args, **kwargs):
        self.batch_size = kwargs.get('batch_size', 4)
        super().__init__(*args, **kwargs)
        self._tokenizer = AutoTokenizer.from_pretrained(self.model, chunk_size=self.max_tokens)
        self._model = AutoModelForCausalLM.from_pretrained(self.model, trust_remote_code=True)
        self._llm = HuggingFacePipeline.from_model_id(
            model_id=self.model,
            task=self.task,
            device_map='auto',
            batch_size=self.batch_size,
            model_kwargs={
                "max_length": self.max_tokens,
                "trust_remote_code": True
            },
            pipeline_kwargs={
                "temperature": self.temperature,
                "repetition_penalty":1.1,
                "max_new_tokens": self.max_tokens,
                **self.args
            }
        )



from abc import ABC
from typing import Any
from aiohttp import web
# Langchain:
from langchain_core.prompts import (
    PromptTemplate,
    ChatPromptTemplate
)
from langchain.agents.agent import (
    AgentExecutor,
    RunnableAgent,
    RunnableMultiActionAgent,
)
from langchain.memory import (
    ConversationBufferMemory
)
from langchain.agents import create_react_agent
from langchain_community.chat_message_histories import (
    RedisChatMessageHistory
)
from navconfig.logging import logging
## LLM configuration
# Vertex
try:
    from ..llms.vertex import VertexLLM
    VERTEX_ENABLED = True
except (ModuleNotFoundError, ImportError):
    VERTEX_ENABLED = False

# Google
try:
    from ..llms.google import GoogleGenAI
    GOOGLE_ENABLED = True
except (ModuleNotFoundError, ImportError):
    GOOGLE_ENABLED = False

# Anthropic:
try:
    from ..llms.anthropic import Anthropic
    ANTHROPIC_ENABLED = True
except (ModuleNotFoundError, ImportError):
    ANTHROPIC_ENABLED = False

# OpenAI
try:
    from ..llms.openai import OpenAILLM
    OPENAI_ENABLED = True
except (ModuleNotFoundError, ImportError):
    OPENAI_ENABLED = False

# Groq
try:
    from ..llms.groq import GroqLLM
    GROQ_ENABLED = True
except (ModuleNotFoundError, ImportError):
    GROQ_ENABLED = False

from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)  # for exponential backoff
from ..interfaces import DBInterface
from ..models import AgentResponse
from ..conf import REDIS_HISTORY_URL


BASEPROMPT = """
Your name is {name}. You are a helpful and advanced AI assistant equipped with various tools to help users find information and solve problems efficiently.
You are designed to be able to assist with a wide range of tasks.
Overall, Assistant is a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics.
Whether you need help with a specific question or just want to have a conversation about a particular topic, Assistant is here to assist.

**Has access to the following tools:**

- {tools}
- Google Web Search: Perform web searches to retrieve the latest and most relevant information from the internet.
- google_maps_location_finder: Find location information, including latitude, longitude, and other geographical details.
- Wikipedia: Access detailed and verified information from Wikipedia.
- WikiData: Fetch structured data from WikiData for precise and factual details.
- Bing Search: Search the web using Microsoft Bing Search, conduct searches using Bing for alternative perspectives and sources.
- DuckDuckGo Web Search: Search the web using DuckDuckGo Search.
- zipcode_distance: Calculate the distance between two zip codes.
- zipcode_location: Obtain geographical information about a specific zip code.
- zipcodes_by_radius: Find all US zip codes within a given radius of a zip code.
- asknews_search: Search for up-to-date news and historical news on AskNews site.
- StackExchangeSearch: Search for questions and answers on Stack Exchange.
- openweather_tool: Get current weather conditions based on specific location, providing latitude and longitude.
- OpenWeatherMap: Get weather information about a location.
- yahoo_finance_news: Retrieve the latest financial news from Yahoo Finance.
- python_repl_ast: A Python shell. Use this to execute python commands. Input should be a valid python command. When using this tool, sometimes output is abbreviated - make sure it does not look abbreviated before using it in your answer.
- youtube_search: Search for videos on YouTube based on specific keywords.


Use these tools effectively to provide accurate and comprehensive responses.

**Instructions:**
1. Understand the Query: Comprehend the user's request, especially if it pertains to events that may have already happened.
2. **Event Timing Validation**: For questions about recent events or events that may have happened already (like sporting events, conferences, etc.), if you're not confident that the event has happened, you must **use one of the web search tools** to confirm before making any conclusions.
3. Determine Confidence: If confident (90%+), provide the answer directly within the Thought process. If not confident, **always use a web search tool**.
4. Choose Tool: If needed, select the most suitable tool, using one of [{tool_names}].
5. Collect Information: Use the tool to gather data.
6. Analyze Information: Identify patterns, relationships, and insights.
7. Synthesize Response: Combine the information into a clear response.
8. Cite Sources: Mention the sources of the information.

** Your Style: **
- Maintain a professional and friendly tone.
- Be clear and concise in your explanations.
- Use simple language for complex topics to ensure user understanding.

To respond directly, use the following format:
```
Question: the input question you must answer.
Thought: Explain your reasoning.
Final Thought: Summarize your findings.
Final Answer: Provide a clear and structured answer to the original question with relevant details.
```


To use a tool, please use the following format:
```
Question: the input question you must answer.
Thought: Explain your reasoning, including whether you need to use a tool.
Action: the action to take, should be one of [{tool_names}].
- If using a tool: Specify the tool name (e.g., "Google Web Search") and the input.
Action Input: the input to the action.
Observation: the result of the action.
... (this Thought/Action/Action Input/Observation can repeat N times)
Final Thought: Summarize your findings.
Final Answer: Provide a clear and structured answer to the original question with relevant details.
Detailed Result: Include the detailed result from the tool here if applicable.
```

**Important**: For recent events (such as the Paris 2024 Olympic Games), you must **use a web search tool** to verify the outcome or provide accurate up-to-date information before concluding. Always prioritize using tools if you're unsure or if the event is recent.


Begin!

Question: {input}
{agent_scratchpad}
"""

class BaseAgent(ABC, DBInterface):
    """Base Agent Interface.

    This is the base class for all Agent Chatbots. It is an abstract class that
    must be implemented by all Agent Chatbots.

    """
    def __init__(
        self,
        name: str = 'Agent',
        llm: str = 'vertexai',
        tools: list = None,
        prompt_template: str = None,
        **kwargs
    ):
        self.name = name
        self.tools = tools
        self.prompt_template = prompt_template
        if not self.prompt_template:
            self.prompt_template = BASEPROMPT
        self.llm = self.llm_chain(llm, **kwargs)
        self.prompt = self.get_prompt(self.prompt_template)
        self.kwargs = kwargs
        # Bot:
        self._agent = None
        self.agent = None
        # Logger:
        self.logger = logging.getLogger('Parrot.Agent')

    def get_prompt(self, prompt, **kwargs):
        partial_prompt = ChatPromptTemplate.from_template(prompt)
        return partial_prompt.partial(
            tools=self.tools,
            name=self.name,
            **kwargs
        )

    def get_memory(
        self,
        session_id: str = None,
        key: str = 'chat_history',
        input_key: str = 'question',
        output_key: str = 'answer',
        size: int = 5,
        ttl: int = 86400
    ):
        args = {
            'memory_key': key,
            'input_key': input_key,
            'output_key': output_key,
            'return_messages': True,
            'max_len': size
        }
        if session_id:
            message_history = RedisChatMessageHistory(
                url=REDIS_HISTORY_URL,
                session_id=session_id,
                ttl=ttl
            )
            args['chat_memory'] = message_history
        return ConversationBufferMemory(
            **args
        )

    def create_agent(self, **kwargs):
        # Create a ReAct Agent:
        self.agent = RunnableAgent(
            runnable = create_react_agent(
                self.llm,
                self.tools,
                prompt=self.prompt,
            ),  # type: ignore
            input_keys_arg=["input"],
            return_keys_arg=["output"],
        )
        return self.agent

    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
    def llm_chain(
        self, llm: str = "vertexai", **kwargs
    ):
        """llm_chain.

        Args:
            llm (str): The language model to use.

        Returns:
            str: The language model to use.

        """
        if llm == 'openai':
            mdl = OpenAILLM(model="gpt-3.5-turbo", **kwargs)
        elif llm in ('vertexai', 'VertexLLM'):
            mdl = VertexLLM(model="gemini-1.5-pro", **kwargs)
        elif llm == 'anthropic':
            mdl = Anthropic(model="claude-3-opus-20240229", **kwargs)
        elif llm in ('groq', 'Groq', 'llama3'):
            mdl = GroqLLM(model="llama3-70b-8192", **kwargs)
        elif llm == 'mixtral':
            mdl = GroqLLM(model="mixtral-8x7b-32768", **kwargs)
        elif llm == 'google':
            mdl = GoogleGenAI(model="models/gemini-1.5-pro-latest", **kwargs)
        else:
            raise ValueError(f"Invalid llm: {llm}")

        # get the LLM:
        return mdl.get_llm()

    def get_executor(self, agent: RunnableAgent, tools: list, verbose: bool = True):
        return AgentExecutor(
            agent=agent,
            tools=tools,
            # callback_manager=callback_manager,
            verbose=verbose,
            return_intermediate_steps=True,
            max_iterations=5,
            max_execution_time=360,
            handle_parsing_errors=True,
            # memory=self.memory,
            # early_stopping_method='generate',
            **(self.kwargs or {}),
        )

    def get_chatbot(self):
        return self.get_executor(self.agent, self.tools)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._agent = None

    def get_conversation(self):
        # Create the agent:
        self.create_agent()
        # define conversation:
        self._agent = self.get_chatbot()
        return self

    def get_retrieval(self, request: web.Request) -> Any:
        # Create the agent:
        self.create_agent()
        # define conversation:
        self._agent = self.get_chatbot()
        self.request = request
        return self

    async def question(
            self,
            question: str = None,
            chain_type: str = 'stuff',
            search_type: str = 'similarity',
            search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
            return_docs: bool = True,
            metric_type: str = None,
            memory: Any = None,
            **kwargs
    ):
        """question.

        Args:
            question (str): The question to ask the chatbot.
            chain_type (str): The type of chain to use.
            search_type (str): The type of search to use.
            search_kwargs (dict): The search kwargs to use.
            return_docs (bool): Return the documents.
            metric_type (str): The metric type to use.
            memory (Any): The memory to use.

        Returns:
            Any: The response from the chatbot.

        """
        # TODO: adding the vector-search to the agent
        input_question = {
            "input": question
        }
        result = self._agent.invoke(input_question)
        try:
            response = AgentResponse(question=question, **result)
            response.response = self.as_markdown(
                response
            )
            return response
        except Exception as e:
            self.logger.exception(
                f"Error on response: {e}"
            )
            raise

    def invoke(self, query: str):
        """invoke.

        Args:
            query (str): The query to ask the chatbot.

        Returns:
            str: The response from the chatbot.

        """
        input_question = {
            "input": query
        }
        result = self._agent.invoke(input_question)
        try:
            response = AgentResponse(question=query, **result)
            try:
                return self.as_markdown(
                    response
                ), response
            except Exception as exc:
                self.logger.exception(
                    f"Error on response: {exc}"
                )
                return result.get('output', None), None
        except Exception as e:
            return result, e

    def as_markdown(self, response: AgentResponse) -> str:
        markdown_output = f"**Question**: {response.question}  \n"
        markdown_output += f"**Answer**: {response.output}  \n"
        markdown_output += "```"
        return markdown_output



from datamodel import BaseModel
from navigator.views import ModelView
from .models import ADUser, ADPeople


class ADUserHandler(ModelView):
    model: BaseModel = ADUser
    pk: str = 'people_id'


class ADPeopleHandler(ModelView):
    model: BaseModel = ADPeople
    pk: str = 'people_id'



import os
import glob
from pathlib import PurePath
from langchain_community.document_loaders import (
    DirectoryLoader
)


def load_directory(
    path: PurePath,
    text_splitter,
    source_type,
    file_pattern: str = "**/*.txt"
):
    """
    Load all Text documents from a directory.

    Args:
        path (str): The path to the directory.
        text_splitter (TextSplitter): A text splitter object.
        source_type (str): The type of source.

    Returns:
        list: A list of documents.
    """
    documents = []
    loader = DirectoryLoader(
        path=str(path),
        glob=file_pattern,
        recursive=True,
        show_progress=True,
        use_multithreading=True
    )
    documents = loader.load()
    for doc in documents:
        doc.metadata['source_type'] = source_type
    return text_splitter.split_documents(documents)



"""Nav Parrot Meta information."""

__title__ = "ai-parrot"
__description__ = "Live Chatbots based on Langchain chatbots and Agents \
    Integrated into Navigator Framework or used into aiohttp applications."
__version__ = "0.3.3"
__author__ = "Jesus Lara"
__author_email__ = "jesuslarag@gmail.com"
__license__ = "MIT"
__copyright__ = 'Copyright (c) 2020-2024 Jesus Lara'



from typing import Union, Optional
import uuid
import time
from datetime import datetime
from pathlib import Path, PurePath
from enum import Enum
from langchain_core.agents import AgentAction

from datamodel import BaseModel, Field
from datamodel.types import Text  # pylint: disable=no-name-in-module
from asyncdb.models import Model

def created_at(*args, **kwargs) -> int:
    return int(time.time()) * 1000


class AgentResponse(BaseModel):
    """AgentResponse.
    dict_keys(
        ['input', 'chat_history', 'output', 'intermediate_steps']
    )

    Response from Chatbots.
    """
    question: str = Field(required=False)
    input: str = Field(required=False)
    output: str = Field(required=False)
    response: str = Field(required=False)
    intermediate_steps: list = Field(default_factory=list)
    chat_history: list = Field(repr=True, default_factory=list)
    source_documents: list = Field(required=False, default_factory=list)

    def __post_init__(self) -> None:
        if self.intermediate_steps:
            steps = []
            for item, result in self.intermediate_steps:
                if isinstance(item, AgentAction):
                    # convert into dictionary:
                    steps.append(
                        {
                            "tool": item.tool,
                            "tool_input": item.tool_input,
                            "result": result,
                            "log": str(item.log)
                        }
                    )
            if steps:
                self.intermediate_steps = steps


class ChatResponse(BaseModel):
    """ChatResponse.
    dict_keys(
        ['question', 'chat_history', 'answer', 'source_documents', 'generated_question']
    )

    Response from Chatbots.
    """
    query: str = Field(required=False)
    result: str = Field(required=False)
    question: str = Field(required=False)
    generated_question: str = Field(required=False)
    answer: str = Field(required=False)
    response: str = Field(required=False)
    chat_history: list = Field(repr=True, default_factory=list)
    source_documents: list = Field(required=False, default_factory=list)
    documents: dict = Field(required=False, default_factory=dict)
    sid: uuid.UUID = Field(primary_key=True, required=False, default=uuid.uuid4)
    at: int = Field(default=created_at)

    def __post_init__(self) -> None:
        if self.result and not self.answer:
            self.answer = self.result
        if self.question and not self.generated_question:
            self.generated_question = self.question
        return super().__post_init__()


# Chatbot Model:
class ChatbotModel(Model):
    """Chatbot.
    CREATE TABLE IF NOT EXISTS navigator.chatbots (
        chatbot_id UUID PRIMARY KEY DEFAULT uuid_generate_v4(),
        name VARCHAR NOT NULL DEFAULT 'Nav',
        description VARCHAR,
        config_file VARCHAR DEFAULT 'config.toml',
        company_information JSONB DEFAULT '{}'::JSONB,
        contact_information VARCHAR,
        contact_form VARCHAR,
        contact_email VARCHAR,
        company_website VARCHAR,
        avatar TEXT,
        enabled BOOLEAN NOT NULL DEFAULT TRUE,
        timezone VARCHAR DEFAULT 'UTC',
        attributes JSONB DEFAULT '{}'::JSONB,
        role VARCHAR DEFAULT 'a Human Resources Assistant',
        goal VARCHAR NOT NULL DEFAULT 'Bring useful information to Users.',
        backstory VARCHAR NOT NULL DEFAULT 'I was created by a team of developers to assist with users tasks.',
        rationale VARCHAR NOT NULL DEFAULT 'Remember to maintain a professional tone. Please provide accurate and relevant information.',
        language VARCHAR DEFAULT 'en',
        template_prompt VARCHAR,
        pre_instructions JSONB DEFAULT '[]'::JSONB,
        llm VARCHAR DEFAULT 'VertexLLM',
        model_name VARCHAR DEFAULT 'gemini-pro',
        model_config JSONB DEFAULT '{}'::JSONB,
        embedding_name VARCHAR DEFAULT 'thenlper/gte-base',
        tokenizer VARCHAR DEFAULT 'thenlper/gte-base',
        summarize_model VARCHAR DEFAULT 'facebook/bart-large-cnn',
        classification_model VARCHAR DEFAULT 'facebook/bart-large-cnn',
        database JSONB DEFAULT '{"vector_database": "MilvusStore", "database": "TROC", "collection_name": "troc_information"}'::JSONB,
        created_at TIMESTAMPTZ DEFAULT NOW(),
        created_by INTEGER,
        updated_at TIMESTAMPTZ DEFAULT NOW()
    );
    """
    chatbot_id: uuid.UUID = Field(primary_key=True, required=False, default_factory=uuid.uuid4)
    name: str = Field(default='Nav', required=True)
    description: str = Field(default='Nav Chatbot', required=False)
    config_file: str = Field(default='config.toml', required=False)
    custom_class: str = Field(required=False)
    company_information: dict = Field(default_factory=dict, required=False)
    avatar: Text
    enabled: bool = Field(required=True, default=True)
    timezone: str = Field(required=False, max=75, default="UTC", repr=False)
    attributes: Optional[dict] = Field(required=False, default_factory=dict)
    # Chatbot Configuration
    role: str = Field(
        default="a Human Resources Assistant",
        required=False
    )
    goal: str = Field(
        default="Bring useful information to Users.",
        required=True
    )
    backstory: str = Field(
        default="I was created by a team of developers to assist with users tasks.",
        required=True
    )
    rationale: str = Field(
        default=(
            "Remember to maintain a professional tone."
            " Please provide accurate and relevant information."
        ),
        required=True
    )
    language: str = Field(default='en', required=False)
    template_prompt: Union[str, PurePath] = Field(
        default=None,
        required=False
    )
    pre_instructions: list = Field(
        default_factory=list,
        required=False
    )
    # Model Configuration:
    llm: str = Field(default='VertexLLM', required=False)
    llm_config: dict = Field(default_factory=dict, required=False)
    embedding_name: str = Field(default="thenlper/gte-base", required=False)
    tokenizer: str = Field(default='thenlper/gte-base', required=False)
    summarize_model: str = Field(default="facebook/bart-large-cnn", required=False)
    classification_model: str = Field(default="facebook/bart-large-cnn", required=False)
    # Database Configuration
    database: dict = Field(default='TROC', required=False, default_factory=dict)
    # Bot/Agent type
    bot_type: str = Field(default='chatbot', required=False)
    # When created
    created_at: datetime = Field(required=False, default=datetime.now())
    created_by: int = Field(required=False)
    updated_at: datetime = Field(required=False, default=datetime.now())

    def __post_init__(self) -> None:
        super(ChatbotModel, self).__post_init__()
        if isinstance(self.config_file, str):
            self.config_file = Path(self.config_file).resolve()
        if isinstance(self.config_file, PurePath):
            self.config_file = str(self.config_file)

    class Meta:
        """Meta Chatbot."""
        driver = 'pg'
        name = "chatbots"
        schema = "navigator"
        strict = True
        frozen = False


class ChatbotUsage(Model):
    """ChatbotUsage.

    Saving information about Chatbot Usage.

    -- ScyllaDB CREATE TABLE Syntax --
    CREATE TABLE IF NOT EXISTS navigator.chatbots_usage (
        chatbot_id TEXT,
        user_id SMALLINT,
        sid TEXT,
        source_path TEXT,
        platform TEXT,
        origin inet,
        user_agent TEXT,
        question TEXT,
        response TEXT,
        used_at BIGINT,
        at TEXT,
        PRIMARY KEY ((chatbot_id, sid, at), used_at)
    ) WITH CLUSTERING ORDER BY (used_at DESC)
    AND default_time_to_live = 10368000;

    """
    chatbot_id: uuid.UUID = Field(primary_key=True, required=False)
    user_id: int = Field(primary_key=True, required=False)
    sid: uuid.UUID = Field(primary_key=True, required=False, default=uuid.uuid4)
    source_path: str = Field(required=False, default='web')
    platform: str = Field(required=False, default='web')
    origin: str = Field(required=False)
    user_agent: str = Field(required=False)
    question: str = Field(required=False)
    response: str = Field(required=False)
    used_at: int = Field(required=False, default=created_at)
    event_timestamp: datetime = Field(required=False, default=datetime.now)
    _at: str = Field(primary_key=True, required=False)

    class Meta:
        """Meta Chatbot."""
        driver = 'bigquery'
        name = "chatbots_usage"
        schema = "navigator"
        ttl = 10368000  # 120 days in seconds
        strict = True
        frozen = False

    def __post_init__(self) -> None:
        if not self._at:
            # Generate a unique session id
            self._at = f'{self.sid}:{self.used_at}'
        super(ChatbotUsage, self).__post_init__()


class FeedbackType(Enum):
    """FeedbackType."""
    # Good Feedback
    GOOD_COMPLETE = "Completeness"
    GOOD_CORRECT = "Correct"
    GOOD_FOLLOW = "Follow the instructions"
    GOOD_UNDERSTAND = "Understandable"
    GOOD_USEFUL = "very useful"
    GOOD_OTHER = "Please Explain"
    # Bad Feedback
    BAD_DONTLIKE = "Don't like the style"
    BAD_INCORRECT = "Incorrect"
    BAD_NOTFOLLOW = "Didn't follow the instructions"
    BAD_LAZY = "Being lazy"
    BAD_NOTUSEFUL = "Not useful"
    BAD_UNSAFE = "Unsafe or problematic"
    BAD_OTHER = "Other"

    @classmethod
    def list_feedback(cls, feedback_category):
        """Return a list of feedback types based on the given category (Good or Bad)."""
        prefix = feedback_category.upper() + "_"
        return [feedback for feedback in cls if feedback.name.startswith(prefix)]

class ChatbotFeedback(Model):
    """ChatbotFeedback.

    Saving information about Chatbot Feedback.

    -- ScyllaDB CREATE TABLE Syntax --
    CREATE TABLE IF NOT EXISTS navigator.chatbots_feedback (
        chatbot_id UUID,
        user_id INT,
        sid UUID,
        at TEXT,
        rating TINYINT,
        like BOOLEAN,
        dislike BOOLEAN,
        feedback_type TEXT,
        feedback TEXT,
        created_at BIGINT,
        PRIMARY KEY ((chatbot_id, user_id, sid), created_at)
    ) WITH CLUSTERING ORDER BY (created_at DESC)
    AND default_time_to_live = 7776000;

    """
    chatbot_id: uuid.UUID = Field(primary_key=True, required=False)
    user_id: int = Field(required=False)
    sid: uuid.UUID = Field(primary_key=True, required=False)
    _at: str = Field(primary_key=True, required=False)
    # feedback information:
    rating: int = Field(required=False, default=0)
    _like: bool = Field(required=False, default=False)
    _dislike: bool = Field(required=False, default=False)
    feedback_type: FeedbackType = Field(required=False)
    feedback: str = Field(required=False)
    created_at: int = Field(required=False, default=created_at)
    expiration_timestamp: datetime = Field(required=False, default=datetime.now)

    class Meta:
        """Meta Chatbot."""
        driver = 'bigquery'
        name = "chatbots_feedback"
        schema = "navigator"
        ttl = 7776000  # 3 months in seconds
        strict = True
        frozen = False

    def __post_init__(self) -> None:
        if not self._at:
            # Generate a unique session id
            if not self.created_at:
                self.created_at = created_at()
            self._at = f'{self.sid}:{self.created_at}'
        super(ChatbotFeedback, self).__post_init__()


## Prompt Library:

class PromptCategory(Enum):
    """
    Prompt Category.

    Categorization of Prompts, as "tech", "tech-or-explain", "idea", "explain", "action", "command", "other".
    """
    TECH = "tech"
    TECH_OR_EXPLAIN = "tech-or-explain"
    IDEA = "idea"
    EXPLAIN = "explain"
    ACTION = "action"
    COMMAND = "command"
    OTHER = "other"

class PromptLibrary(Model):
    """PromptLibrary.

    Saving information about Prompt Library.

    -- PostgreSQL CREATE TABLE Syntax --
    CREATE TABLE IF NOT EXISTS navigator.prompt_library (
            prompt_id UUID PRIMARY KEY DEFAULT uuid_generate_v4(),
            chatbot_id UUID,
            title varchar,
            query varchar,
            description TEXT,
            prompt_category varchar,
            prompt_tags varchar[],
            created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
            created_by INTEGER,
            updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
    );
    """
    prompt_id: uuid.UUID = Field(primary_key=True, required=False, default_factory=uuid.uuid4)
    chatbot_id: uuid.UUID = Field(required=True)
    title: str = Field(required=True)
    query: str = Field(required=True)
    description: str = Field(required=False)
    prompt_category: str = Field(required=False, default=PromptCategory.OTHER)
    prompt_tags: list = Field(required=False, default_factory=list)
    created_at: datetime = Field(required=False, default=datetime.now)
    created_by: int = Field(required=False)
    updated_at: datetime = Field(required=False, default=datetime.now)

    class Meta:
        """Meta Prompt Library."""
        driver = 'pg'
        name = "prompt_library"
        schema = "navigator"
        strict = True
        frozen = False


    def __post_init__(self) -> None:
        super(PromptLibrary, self).__post_init__()



from collections.abc import Callable
from typing import Any, Optional, List, Union
from pathlib import PurePath, Path
from io import StringIO
import fitz  # PyMuPDF
from langchain.docstore.document import Document
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter
)
from .basepdf import BasePDF


class PDFFnLoader(BasePDF):
    """
    Loading a PDF with including function processing.
    """
    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Union[None, Callable[..., Any]] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        **kwargs
    ):
        table_settings = kwargs.pop('table_settings', {})
        super().__init__(
            path=path,
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            language=language,
            **kwargs
        )
        if not text_splitter:
            self.text_splitter = RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
                self.tokenizer,
                chunk_size=2000,
                chunk_overlap=100,
                add_start_index=True,  # If `True`, includes chunk's start index in metadata
                strip_whitespace=True,  # strips whitespace from the start and end
                separators=["\n\n", "\n", "\r\n", "\r", "\f", "\v", "\x0b", "\x0c"],
            )
        self.table_settings = {
            "vertical_strategy": "lines",
            "horizontal_strategy": "text",
            "intersection_x_tolerance": 5,
            "intersection_y_tolerance": 5,
            "edge_min_length": 10,
        }
        # Define settings for Fitz Table Processing
        self.table_settings = {**self.table_settings, **table_settings}

    def set_metadata(self, path, page, page_number, **kwargs) -> dict:
        n = page_number + 1
        return {
            "url": '',
            "index": f"{path.name} #{page_number}",
            "source": f"{path.name} #{page_number}",
            "filename": path.name,
            "source_type": self._source_type,
            "type": "pdf",
            "question": "",
            "answer": "",
            "summary": '',
            "document_meta": {
                "page_number": n,
                **kwargs
            }
        }

    def processing_table(self, table, table_idx, page, **kwargs) -> dict:
        df = table.to_pandas()  # convert to pandas DataFrame
        df = df.dropna(axis=1, how='all')
        df = df.dropna(how='all', axis=0)  # Drop empty rows
        table_data = []
        # Extract text from each cell
        for row_idx in range(table.row_count):
            for col_idx in range(table.column_count):
                cell = table[row_idx][col_idx]
                print('CELL ', cell)
                print('---------')
                cell_text = cell.get_text("text", flags=fitz.TEXTFLAGS_HTML)
                print(cell_text)

        return table_data

    def _load_pdf(self, path: PurePath, **kwargs):
        """
        Open a PDF file.

        Args:
            path (PurePath): The path to the PDF file.

        Returns:
            fitz.PDF: The PDF object.
        """
        pdf = fitz.open(path)
        self.logger.info(f"Loading PDF file: {path}")
        for page_num in range(len(pdf)):
            # Will extract first the table and second the block of texts
            page = pdf.load_page(page_num)
            parts = page.get_text("dict", flags=fitz.TEXTFLAGS_HTML)
            # print('PARTS ', parts)
            blocks = page.get_text("dict")["blocks"]
            # print('BLOCKS >', blocks)
            metadata = self.set_metadata(path, page, page_num)
            # print('META > ', metadata)
            tables = page.find_tables(**self.table_settings)
            for tab_idx, table in enumerate(tables):
                table_data = self.processing_table(table, tab_idx, page)
        return []



from typing import Any
from qdrant_client import QdrantClient  # pylint: disable=import-error
from langchain_community.vectorstores import (  # pylint: disable=import-error, E0611
    Qdrant
)
from .abstract import AbstractStore
from ..conf import (
    QDRANT_PROTOCOL,
    QDRANT_HOST,
    QDRANT_PORT,
    QDRANT_USE_HTTPS,
    QDRANT_CONN_TYPE,
    QDRANT_URL
)


class QdrantStore(AbstractStore):
    """QdrantStore class.


    Args:
        host (str): Qdrant host.
        port (int): Qdrant port.
        index_name (str): Qdrant index name.
    """

    def _create_qdrant_client(self, host, port, url, https, verify, qdrant_args):
        """
        Creates a Qdrant client based on the provided configuration.

        Args:
            host: Host of the Qdrant server (if using "server" connection).
            port: Port of the Qdrant server (if using "server" connection).
            url: URL of the Qdrant cloud service (if using "cloud" connection).
            https: Whether to use HTTPS for the connection.
            verify: Whether to verify the SSL certificate.
            qdrant_args: Additional arguments for the Qdrant client.

        Returns:
            A QdrantClient object.
        """
        if url is not None:
            return QdrantClient(
            url=url,
            port=None,
            verify=verify,
            **qdrant_args
            )
        else:
            return QdrantClient(
            host,
            port=port,
            https=https,
            verify=verify,
            **qdrant_args
            )

    def __init__(self, embeddings = None, **kwargs):
        super().__init__(embeddings, **kwargs)
        self.host = kwargs.get("host", QDRANT_HOST)
        self.port = kwargs.get("port", QDRANT_PORT)
        qdrant_args = kwargs.get("qdrant_args", {})
        connection_type = kwargs.get("connection_type", QDRANT_CONN_TYPE)
        url = kwargs.get("url", QDRANT_URL)
        if connection_type == "server":
            self.client = self._create_qdrant_client(
                self.host, self.port, url, QDRANT_USE_HTTPS, False, qdrant_args
            )
        elif connection_type == "cloud":
            if url is None:
                raise ValueError(
                    "A URL is required for 'cloud' connection"
                )
            self.client = self._create_qdrant_client(
                None, None, url, False, False, qdrant_args
            )
        else:
            raise ValueError(
                f"Invalid connection type: {connection_type}"
            )
        if url is not None:
            self.url = url
        else:
            self.url = f"{QDRANT_PROTOCOL}://{self.host}"
            if self.port:
                self.url += f":{self.port}"

    def get_vectorstore(self):
        if self._embed_ is None:
            _embed_ = self.create_embedding(
                model_name=self.embedding_name
            )
        else:
            _embed_ = self._embed_
        self.vector = Qdrant(
            client=self.client,
            collection_name=self.collection,
            embeddings=_embed_,
        )
        return self.vector

    async def load_documents(
        self,
        documents: list,
        collection: str = None
    ):
        if collection is None:
            collection = self.collection

        docstore = Qdrant.from_documents(
            documents,
            self._embed_,
            url=self.url,
            # location=":memory:",  # Local mode with in-memory storage only
            collection_name=collection,
            force_recreate=False,
        )
        return docstore

    def upsert(self, payload: dict, collection: str = None) -> None:
        if collection is None:
            collection = self.collection
        self.client.upsert(
            collection_name=collection,
            points=self._embed_,
            payload=payload
        )

    def search(self, payload: dict, collection: str = None) -> dict:
        pass

    async def delete_collection(self, collection: str = None) -> dict:
        self.client.delete_collection(
            collection_name=collection
        )

    async def create_collection(
        self,
        collection_name: str,
        document: Any,
        dimension: int = 768,
        **kwargs
    ) -> dict:
        # Here using drop_old=True to force recreate based on the first document
        docstore = Qdrant.from_documents(
            [document],
            self._embed_,
            url=self.url,
            # location=":memory:",  # Local mode with in-memory storage only
            collection_name=collection_name,
            force_recreate=True,
        )
        return docstore



from parrot.chatbots import Chatbot

class HRAgent(Chatbot):
    """Represents an Human Resources agent in Navigator.

        Each agent has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    name: str = 'TROCers'
    company: str = 'T-ROC Global'
    company_website: str = 'https://www.trocglobal.com'
    contact_information = 'communications@trocglobal.com'
    contact_form = 'https://www.surveymonkey.com/r/TROC_Suggestion_Box'
    role: str = 'Human Resources Assistant'
    goal = 'Bring useful information to employees'



from collections.abc import Callable
from typing import Any, Union, List
from abc import abstractmethod
from pathlib import Path
from moviepy.editor import VideoFileClip
from transformers import (
    pipeline,
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    GenerationConfig
)
from transformers import (
    WhisperForConditionalGeneration,
    WhisperProcessor,
    WhisperTimeStampLogitsProcessor
)

from langchain.chains.summarize import load_summarize_chain
from langchain.text_splitter import (
    TokenTextSplitter,
)
from .abstract import AbstractLoader


def extract_video_id(url):
    parts = url.split("?v=")
    video_id = parts[1].split("&")[0]
    return video_id


class BaseVideoLoader(AbstractLoader):
    """
    Generating Video transcripts from Videos.
    """
    _extension = ['.youtube']
    encoding = 'utf-8'
    chunk_size = 768

    def __init__(
        self,
        urls: List[str],
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'video',
        language: str = "en",
        video_path: Union[str, Path] = None,
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type, **kwargs)
        self.urls = urls
        self._task = kwargs.get('task', "automatic-speech-recognition")
        # Topics:
        self.topics: list = kwargs.get('topics', [])
        self._model_size: str = kwargs.get('model_size', 'medium')
        self.summarization_model = "facebook/bart-large-cnn"
        self._model_name: str = kwargs.get('model_name', 'whisper')
        self.summarizer = pipeline(
            "summarization",
            tokenizer=AutoTokenizer.from_pretrained(
                self.summarization_model
            ),
            model=AutoModelForSeq2SeqLM.from_pretrained(
                self.summarization_model
            ),
            device=self._get_device()
        )
        # language:
        self._language = language
        # directory:
        if isinstance(video_path, str):
            self._video_path = Path(video_path).resolve()
        self._video_path = video_path

    def transcript_to_vtt(self, transcript: str, transcript_path: Path) -> str:
        """
        Convert a transcript to VTT format.
        """
        vtt = "WEBVTT\n\n"
        for i, chunk in enumerate(transcript['chunks'], start=1):
            start, end = chunk['timestamp']
            text = chunk['text'].replace("\n", " ")  # Replace newlines in text with spaces

            if start is None or end is None:
                print(f"Warning: Missing timestamp for chunk {i}, skipping this chunk.")
                continue

            # Convert timestamps to WebVTT format (HH:MM:SS.MMM)
            start_vtt = f"{int(start // 3600):02}:{int(start % 3600 // 60):02}:{int(start % 60):02}.{int(start * 1000 % 1000):03}"
            end_vtt = f"{int(end // 3600):02}:{int(end % 3600 // 60):02}:{int(end % 60):02}.{int(end * 1000 % 1000):03}"

            vtt += f"{i}\n{start_vtt} --> {end_vtt}\n{text}\n\n"
        # Save the VTT file
        try:
            with open(str(transcript_path), "w") as f:
                f.write(vtt)
            print(f'Saved VTT File on {transcript_path}')
        except Exception as exc:
            print(f"Error saving VTT file: {exc}")
        return vtt

    def format_timestamp(self, seconds):
        # This helper function takes the total seconds and formats it into hh:mm:ss,ms
        hours, remainder = divmod(int(seconds), 3600)
        minutes, seconds = divmod(remainder, 60)
        milliseconds = int((seconds % 1) * 1000)
        seconds = int(seconds)
        return f"{hours:02}:{minutes:02}:{seconds:02},{milliseconds:03}"

    def transcript_to_blocks(self, transcript: str) -> list:
        """
        Convert a transcript to blocks.
        """
        blocks = []
        for i, chunk in enumerate(transcript['chunks'], start=1):
            current_window = {}
            start, end = chunk['timestamp']
            if start is None or end is None:
                print(f"Warning: Missing timestamp for chunk {i}, skipping this chunk.")
                continue

            start_srt = self.format_timestamp(start)
            end_srt = self.format_timestamp(end)
            text = chunk['text'].replace("\n", " ")  # Replace newlines in text with spaces
            current_window['id'] = i
            current_window['start_time'] = start_srt
            current_window['end_time'] = end_srt
            current_window['text'] = text
            blocks.append(current_window)
        return blocks

    def transcript_to_srt(self, transcript: str) -> str:
        """
        Convert a transcript to SRT format.
        """
        # lines = transcript.split("\n")
        srt = ""
        for i, chunk in enumerate(transcript['chunks'], start=1):
            start, end = chunk['timestamp']
            text = chunk['text'].replace("\n", " ")  # Replace newlines in text with spaces
            # Convert start and end times to SRT format HH:MM:SS,MS
            start_srt = f"{start // 3600:02}:{start % 3600 // 60:02}:{start % 60:02},{int(start * 1000 % 1000):03}"
            end_srt = f"{end // 3600:02}:{end % 3600 // 60:02}:{end % 60:02},{int(end * 1000 % 1000):03}"
            srt += f"{i}\n{start_srt} --> {end_srt}\n{text}\n\n"
        return srt

    def chunk_text(self, text, chunk_size, tokenizer):
        # Tokenize the text and get the number of tokens
        tokens = tokenizer.tokenize(text)
        # Split the tokens into chunks
        for i in range(0, len(tokens), chunk_size):
            yield tokenizer.convert_tokens_to_string(
                tokens[i:i+chunk_size]
            )

    def get_summary(self, documents: list) -> str:
        """
        Get a summary of a text.
        """
        try:
            splitter = TokenTextSplitter(
                chunk_size=5000,
                chunk_overlap=100,
            )
            summarize_chain = load_summarize_chain(
                llm=self._llm,
                chain_type="refine"
            )
            chunks = splitter.split_documents(documents)
            summary = summarize_chain.invoke(chunks)
            return summary
        except Exception as e:
            print('ERROR in get_summary:', e)
            return ""

    def summarization(self, text: str) -> str:
        """
        Get a summary of a text considering token limits.
        """
        try:
            tokenizer = self.summarizer.tokenizer
            # to be safe under the limit
            max_length = tokenizer.model_max_length - 10
            summaries = []
            for text_chunk in self.chunk_text(text, max_length, tokenizer):
                chunk_summary = self.summarizer(
                    text_chunk,
                    max_length=150,
                    min_length=30,
                    do_sample=False)[0]['summary_text']
                summaries.append(chunk_summary)
            return " ".join(summaries)
        except Exception as e:
            print('ERROR in summarization:', e)
            return ""

    def extract_audio(self, video_path, audio_path):
        """
        Extracts the audio from a video file and saves it as an audio file.

        Args:
            video_path (str): Path to the video file.
            audio_path (str): Path where the extracted audio file will be saved.
        """
        if audio_path.exists():
            print(f"Audio already extracted: {audio_path}")
            return
        video_clip = VideoFileClip(str(video_path))
        audio_clip = video_clip.audio
        if not audio_clip:
            return
        audio_clip.write_audiofile(str(audio_path))
        audio_clip.close()
        video_clip.close()

    def get_whisper_transcript(self, audio_path: Path, chunk_length: int = 30):
        # Initialize the Whisper parser
        if self._model_name == 'whisper':
            if self._language == 'en':
                model_name = f"openai/whisper-{self._model_size}.en"
            elif self._language == 'es':
                model_name = f"juancopi81/whisper-{self._model_size}-es"
            else:
                model_name = "openai/whisper-large-v3"
        else:
            model_name = self._model_name

        # Load the model and processor
        model = WhisperForConditionalGeneration.from_pretrained(model_name)
        processor = WhisperProcessor.from_pretrained(model_name)

        # Try to load the generation config, fallback to default if it doesn't exist
        try:
            generation_config = GenerationConfig.from_pretrained(model_name)
        except EnvironmentError:
            print(
                f"Warning: No generation_config.json found for model {model_name}. Using default configuration."
            )
            generation_config = model.generation_config

        # Check and set the no_timestamps_token_id if it doesn't exist
        if not hasattr(model.config, 'no_timestamps_token_id'):
            model.config.no_timestamps_token_id = processor.tokenizer.convert_tokens_to_ids('<|notimestamps|>')

        # Define the generation configuration with WhisperTimeStampLogitsProcessor
        try:
            model.config.logits_processor = [
                WhisperTimeStampLogitsProcessor(generation_config)
            ]
        except Exception:
            model.config.logits_processor = [
                WhisperTimeStampLogitsProcessor(model.config)
            ]

        whisper_pipe = pipeline(
            task=self._task,
            model=model,
            tokenizer=processor.tokenizer,
            feature_extractor=processor.feature_extractor,
            device=self._get_device(),
            chunk_length_s=chunk_length,
            use_fast=True
        )
        if audio_path.exists() and audio_path.stat().st_size > 0:
            # Use the parser to extract transcript
            return whisper_pipe(
                str(audio_path),
                return_timestamps=True
            )
        return None

    @abstractmethod
    def load(self) -> list:
        pass

    @abstractmethod
    def load_video(self, url: str, video_title: str, transcript: str) -> list:
        pass

    def parse(self, source):
        pass



from collections.abc import Callable
from pathlib import Path, PurePath
from typing import Any
import fitz
from pdf4llm import to_markdown
from PIL import Image
from langchain.docstore.document import Document
from langchain.text_splitter import MarkdownTextSplitter
from transformers import (
    AutoTokenizer,
    AutoProcessor,
    LlavaForConditionalGeneration,
    pipeline,
    BitsAndBytesConfig
)
import torch
from .basepdf import BasePDF


quantization_config = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_compute_dtype=torch.float16
)


class PDFImageLoader(BasePDF):
    """
    Loader for PDF files.
    """
    default_prompt: str = "<|user|>\n<image>\nExplain this schematic diagram or technical installation instructions and wire diagrams, please be detailed about descriptions of steps:<|end|>\n<|assistant|>\n"
    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        **kwargs
    ):
        super().__init__(
            path=path,
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            language=language,
            **kwargs
        )
        self._image_model = kwargs.get('image_model', 'llava-hf/llava-v1.6-vicuna-7b-hf')
        self._task = kwargs.get('task', 'image-to-text')
        self._max_tokens = kwargs.get('max_tokens', 600)
        # Loading the model with low CPU memory usage
        # model = LlavaForConditionalGeneration.from_pretrained(
        #     self._image_model,
        #     quantization_config=quantization_config,
        #     device_map="auto",
        #     torch_dtype=torch.float16,
        #     low_cpu_mem_usage=True
        # )
        # # Load the processor
        # processor = AutoProcessor.from_pretrained(self._image_model, use_fast=True)
        self._pipeline = pipeline(
            self._task,
            model=self._image_model,
            # tokenizer=processor.tokenizer,
            # image_processor=processor.image_processor,
            model_kwargs={"quantization_config": quantization_config},
            # device=self._device,
            max_new_tokens=self._max_tokens,
            # low_cpu_mem_usage=True,
            use_fast=True
        )
        # default prompt
        self._prompt = kwargs.get('prompt', self.default_prompt)
        # Markdown Splitter
        self._splitter = MarkdownTextSplitter(
            chunk_size = self._chunk_size,
            chunk_overlap=10
        )

    def pixmap_to_pil_image(self, pix):
        """Converts a PyMuPDF Pixmap object to a PIL Image"""
        return Image.frombytes(
            "RGB",
            [pix.width, pix.height],
            pix.samples
        )

    def _load_pdf(self, path: Path) -> list:
        """
        Load a PDF file as Images.

        Args:
            path (Path): The path to the PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading PDF file: {path}")
            pdf = fitz.open(str(path))  # Open the PDF file
            docs = []
            try:
                # get markdown for all pages and saved separately
                md_text = to_markdown(pdf)
                try:
                    summary = self.get_summary_from_text(md_text)
                except Exception:
                    summary = ''
                metadata = {
                    "url": '',
                    "idx": str(path.name),
                    "filename": str(path.name),
                    "source": str(path.name),
                    "type": 'pdf',
                    "question": '',
                    "answer": '',
                    "data": {},
                    "summary": summary,
                    "source_type": self._source_type,
                    "document_meta": {
                        "title": pdf.metadata.get("title", ""),
                        "author": pdf.metadata.get("author", ""),
                    }
                }
                for idx, chunk in enumerate(self._splitter.split_text(md_text)):
                    _info = {
                        "index": f"{idx}",
                        **metadata
                    }
                    docs.append(
                        Document(
                            page_content=chunk,
                            metadata=_info
                        )
                    )
            except (IndexError, ValueError) as exc:
                self.logger.warning(
                    f"There is no text data to load on {path.name}: {exc}"
                )
            # Then, processing the pages one by one as Images:
            file_name = path.stem.replace(" ", "_").replace(".", "_")
            for page_number in range(pdf.page_count):
                page_num = page_number + 1
                self.logger.notice(
                    f"Processing PDF {path} on Page {page_num}"
                )
                page = pdf[page_number]
                pix = page.get_pixmap(colorspace=fitz.csRGB, alpha=False)
                zoom_x = 2.0  # horizontal zoom
                zoom_y = 2.0  # vertical zoom
                mat = fitz.Matrix(zoom_x, zoom_y)  # zoom factor 2 in each dimension
                img_stream = self.pixmap_to_pil_image(pix)
                url = ''
                img_name = f'image_{file_name}_{page_num}.png'
                if self.save_images is True:
                    img_path = self.save_image(
                        img_stream,
                        img_name,
                        self._imgdir
                    )
                    url = f'/static/images/{img_name}'
                # extracting features and explanations:
                outputs = self._pipeline(
                    img_stream,
                    prompt=self._prompt,
                    generate_kwargs={"max_new_tokens": self._max_tokens}
                )
                documents = []
                for idx, output in enumerate(outputs):
                    generated_text = output['generated_text']
                    # Split using the special tokens, if available
                    split_text = generated_text.split("<|assistant|>")
                    prompt_text = split_text[0].replace("<|prompt|>", "").strip() if "<|prompt|>" in generated_text else ""
                    response_text = split_text[1].strip() if len(split_text) > 1 else ""
                    # Attach the image using Markdown syntax
                    image_markdown = f"\n\n![Image]({url})\n"
                    response_text += image_markdown
                    _meta = {
                        "url": f"{url}",
                        "filename": str(path.name),
                        "index": f"Page {page_num}, part: {idx}",
                        "source": str(path.name),
                        "type": 'pdf',
                        "question": prompt_text,
                        "answer": '',
                        "data": {},
                        "summary": '',
                        "source_type": self._source_type,
                        "document_meta": {
                            "page": f"Page {page}",
                            "image": f"{img_name}",
                            "url": f"{url}"
                        }
                    }
                    documents.append(
                        Document(
                            page_content=response_text,
                            metadata=_meta
                        )
                    )
            return docs + documents

    def load(self) -> list:
        try:
            return super().load()
        finally:
            self._pipeline = None



from langchain_community.tools.google_trends import GoogleTrendsQueryRun
from langchain_community.utilities.google_trends import GoogleTrendsAPIWrapper
from crewai_tools import BaseTool


class GoogleTrendsTool(BaseTool):
    """Google Trends Tool."""
    name: str = "Google Trends"
    description: str = "Search the web using Google Trends API"

    def _run(self, query: str) -> dict:
        """Run the Google Trends Tool."""
        google_trends = GoogleTrendsQueryRun(
            api_wrapper=GoogleTrendsAPIWrapper()
        )
        return google_trends.run(query)



from abc import ABC, abstractmethod
from typing import Union, Any
from collections.abc import Callable
import torch
from langchain_huggingface import (
    HuggingFaceEmbeddings
)
from langchain_community.embeddings import (
    HuggingFaceBgeEmbeddings
)
from langchain_community.embeddings.fastembed import FastEmbedEmbeddings
from navconfig.logging import logging
from ..conf import (
    EMBEDDING_DEVICE,
    EMBEDDING_DEFAULT_MODEL,
    CUDA_DEFAULT_DEVICE,
    MAX_BATCH_SIZE
)


class AbstractStore(ABC):
    """AbstractStore class.

    Args:
        embeddings (str): Embeddings.
    """

    def __init__(self, embeddings: Union[str, Callable] = None, **kwargs):
        self.client: Callable = None
        self.vector: Callable = None
        self._embed_: Callable = None
        self._connected: bool = False
        self.use_bge: bool = kwargs.pop("use_bge", False)
        self.fastembed: bool = kwargs.pop("use_fastembed", False)
        self.embedding_name: str = kwargs.pop('embedding_name', EMBEDDING_DEFAULT_MODEL)
        self.dimension: int = kwargs.pop("dimension", 768)
        self._metric_type: str = kwargs.pop("metric_type", 'COSINE')
        self._index_type: str = kwargs.pop("index_type", 'IVF_FLAT')
        self.database: str = kwargs.pop('database', '')
        self.collection = kwargs.pop("collection_name", "my_collection")
        self.index_name = kwargs.pop("index_name", "my_index")
        if embeddings is not None:
            if isinstance(embeddings, str):
                self.embedding_name = embeddings
            else:
                self._embed_ = embeddings
        self.logger = logging.getLogger(f"Store.{__name__}")
        # client
        self._client = None
        self._client_id = None

    @property
    def connected(self) -> bool:
        return self._connected

    async def __aenter__(self):
        try:
            self.tensor = torch.randn(1000, 1000).cuda()
        except RuntimeError:
            self.tensor = None
        if self._embed_ is None:
            self._embed_ = self.create_embedding(
                model_name=self.embedding_name
            )
        self._client, self._client_id = self.connect()
        return self

    @abstractmethod
    def connect(self):
        pass

    def __enter__(self):
        if self._embed_ is None:
            self._embed_ = self.create_embedding(
                model_name=self.embedding_name
            )
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        # closing Embedding
        self._embed_ = None
        del self.tensor
        try:
            torch.cuda.empty_cache()
        except RuntimeError:
            pass

    def __exit__(self, exc_type, exc_value, traceback):
        # closing Embedding
        self._embed_ = None
        try:
            torch.cuda.empty_cache()
        except RuntimeError:
            pass

    @abstractmethod
    def get_vector(self):
        pass

    @abstractmethod
    async def load_documents(
        self,
        documents: list,
        collection: str = None
    ):
        pass

    @abstractmethod
    def upsert(self, payload: dict, collection_name: str = None) -> None:
        pass

    @abstractmethod
    def search(self, payload: dict, collection_name: str = None) -> dict:
        pass

    @abstractmethod
    async def delete_collection(self, collection_name: str = None) -> dict:
        pass

    @abstractmethod
    async def create_collection(self, collection_name: str, document: Any) -> dict:
        pass

    def create_embedding(
        self,
        model_name: str = None
    ):
        encode_kwargs: str = {
            'normalize_embeddings': True,
            "batch_size": MAX_BATCH_SIZE
        }
        if torch.backends.mps.is_available():
            # Use CUDA Multi-Processing Service if available
            device = torch.device("mps")
        elif torch.cuda.is_available():
            # Use CUDA GPU if available
            device = torch.device(
                f'cuda:{CUDA_DEFAULT_DEVICE}'
            )
        elif EMBEDDING_DEVICE == 'cuda':
            device = torch.device(
                f'cuda:{CUDA_DEFAULT_DEVICE}'
            )
        else:
            device = torch.device(EMBEDDING_DEVICE)
        model_kwargs: str = {'device': device}
        if model_name is None:
            model_name = EMBEDDING_DEFAULT_MODEL
        if self.use_bge is True:
            return HuggingFaceBgeEmbeddings(
                model_name=model_name,
                model_kwargs=model_kwargs,
                encode_kwargs=encode_kwargs
            )
        if self.fastembed is True:
            return FastEmbedEmbeddings(
                model_name=model_name,
                max_length=1024,
                threads=4
            )
        return HuggingFaceEmbeddings(
            model_name=model_name,
            model_kwargs=model_kwargs,
            encode_kwargs=encode_kwargs
        )

    def get_default_embedding(
        self,
        model_name: str = EMBEDDING_DEFAULT_MODEL
    ):
        return self.create_embedding(model_name=model_name)



from pathlib import PurePath
from langchain_community.document_loaders.csv_loader import CSVLoader as CSVL
from .abstract import AbstractLoader


class CSVLoader(AbstractLoader):
    """
    Loader for CSV files.
    """
    _extension = ['.csv']
    csv_args: dict = {
        "delimiter": ",",
        "quotechar": '"',
        "escapechar": "\\",
        "skipinitialspace": False,
        "lineterminator": "\n",
        "quoting": 0,
        "skiprows": 0,
        "encoding": None
    }

    def load(self, path: PurePath) -> list:
        """
        Load data from a CSV file.

        Args:
            source (str): The path to the CSV file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading CSV file: {path}")
            loader = CSVL(
                file_path=path,
                csv_args=self.csv_args,
                autodetect_encoding=True
            )
            documents = loader.load()
            return self.split_documents(documents)
        else:
            return []



from bs4 import BeautifulSoup
from markdownify import MarkdownConverter
from webdriver_manager.chrome import ChromeDriverManager
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from langchain.docstore.document import Document
from langchain.text_splitter import MarkdownTextSplitter
from navconfig.logging import logging
from .abstract import AbstractLoader


logging.getLogger(name='selenium.webdriver').setLevel(logging.WARNING)
logging.getLogger(name='WDM').setLevel(logging.WARNING)
logging.getLogger(name='matplotlib').setLevel(logging.WARNING)


class WebLoader(AbstractLoader):
    """Class to load web pages and extract text."""
    chrome_options = [
        "--headless",
        "--enable-automation",
        "--lang=en",
        "--disable-extensions",
        "--disable-gpu",
        "--no-sandbox",
        "--disable-features=NetworkService",
        "--disable-dev-shm-usage",
        "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"
    ]
    def __init__(self, urls: dict, source_type: str = 'website', **kwargs):
        self.urls = urls
        self._source_type = source_type
        self._options = Options()
        self.timeout: int = kwargs.pop('timeout', 60)
        for option in self.chrome_options:
            self._options.add_argument(option)
        self.driver = webdriver.Chrome(
            service=Service(ChromeDriverManager().install()),
            options=self._options
        )
        self._mark_splitter = MarkdownTextSplitter(
            chunk_size = 1024,
            chunk_overlap=10
        )
        super().__init__(source_type=source_type, **kwargs)

    def md(self, soup, **options):
        return MarkdownConverter(**options).convert_soup(soup)

    def clean_html(self, html, tags, objects=[]):
        soup = BeautifulSoup(html, 'html.parser')
        page_title = soup.title.string
        md_text = self.md(soup)
        # Remove script and style elements
        for script_or_style in soup(["script", "style", "link"]):
            script_or_style.decompose()
        # Extract Content
        content = []
        paragraphs = [' '.join(p.get_text().split()) for p in soup.find_all(tags)]
        # Look for iframe elements and format their src attributes into readable strings
        iframes = soup.find_all('iframe')
        for iframe in iframes:
            video_src = iframe.get('src', '')
            # You might want to customize the formatting of this string
            formatted_video = f"Video Link: {video_src}" if video_src else ""
            content.append(formatted_video)
        if objects:
            for obj in objects:
                (element, args), = obj.items()
                if 'parse_list' in args:
                    parse_list = args.pop('parse_list')
                    # Find the element container:
                    container = soup.find(element, attrs=args)
                    # Parse list of objects (UL, LI)
                    name_type = parse_list.pop('type')
                    params = parse_list.get('find')
                    el = params.pop(0)
                    try:
                        attrs = params.pop(0)
                    except IndexError:
                        attrs = {}
                    elements = container.find_all(el, attrs=attrs)
                    structured_text = ''
                    for element in elements:
                        title = element.find('span', class_='title').get_text(strip=True)
                        lists = element.find_all('ul')
                        if lists:
                            structured_text += f"\nCategory: {title}\n{name_type}:\n"
                        for ul in lists:
                            items = [f"- {li.get_text(strip=True)}" for li in ul.select('li')]
                            formatted_list = '\n'.join(items)
                            structured_text += formatted_list
                        structured_text += "\n"
                    content.append(structured_text)
                else:
                    elements = soup.find_all(element, attrs=args)
                    for element in elements:
                        # Handle <a> tags within the current element
                        links = element.find_all('a')
                        for link in links:
                            # Extract link text and href, format them into a readable string
                            link_text = link.get_text(strip=True)
                            href = link.get('href', '')
                            formatted_link = (
                                f"{link_text} (Link: {href})"
                                if href
                                else link_text
                            )
                            # Replace the original link text in the element
                            # with the formatted version
                            link.replace_with(formatted_link)
                        # work with UL lists:
                        lists = element.find_all('ul')
                        for ul in lists:
                            items = [li.get_text(strip=True) for li in ul.select('li')]
                            formatted_list = '\n'.join(items)
                            content.append(formatted_list)
                        cleaned_text = ' '.join(element.get_text().split())
                        content.append(cleaned_text)
        return (content + paragraphs, md_text, page_title)

    def get(self, address: dict) -> list:
        (url, args), = address.items()
        self.logger.info(
            f'Downloading URL {url} with args {args}'
        )
        locator = args.get('locator', (By.TAG_NAME, 'body'))
        wait = WebDriverWait(self.driver, self.timeout)
        acookies = args.get('accept_cookies', False)
        try:
            self.driver.get(url)
            # After loading page, accept cookies
            wait.until(
                EC.presence_of_element_located(
                    locator
                )
            )
            if acookies:
                btn = wait.until(
                    EC.element_to_be_clickable(
                        acookies
                    )
                )
                btn.click()
        except Exception as exc:
            print(f"Failed to Get {url}: {exc}")
            self.logger.exception(
                str(exc), stack_info=True
            )
            raise
        try:
            extract = args.get('tags', ['p', 'title', 'h1', 'h2', 'section', 'article'])
            objects = args.get('objects', [])
            source_type = args.get('source_type', self._source_type)
            html_content = self.driver.page_source
            content, md_text, page_title = self.clean_html(
                html_content,
                extract,
                objects
            )
            metadata = {
                "source": url,
                "index": page_title,
                "url": url,
                "filename": page_title,
                "question": '',
                "answer": '',
                "source_type": source_type,
                'type': 'webpage',
                'summary': '',
                "document_meta": {
                    "language": "en",
                    "title": page_title,
                },
            }
            docs = []
            if md_text:
                docs.append(
                    Document(
                        page_content=md_text,
                        metadata=metadata
                    )
                )
                # for chunk in self._mark_splitter.split_text(md_text):
                #     docs.append(
                #         Document(
                #             page_content=chunk,
                #             metadata=metadata
                #         )
                #     )
            if content:
                site_content = [
                    Document(
                        page_content=paragraph,
                        metadata=metadata
                    ) for paragraph in content
                ]
            return docs + site_content
        except Exception as exc:
            print(f"Failed to load {url}: {exc}")

    def load(self, **kwargs) -> list:
        documents = []
        for address in self.urls:
            docs = self.get(address)
            if docs:
                documents.extend(docs)
        self.driver.quit()
        return self.split_by_tokens(documents)

    def parse(self, source):
        pass



import streamlit as st
from langchain_community.llms import OpenAI



st.set_page_config(page_title="🦜🔗 Chatbot")
st.title('🦜🔗 Test Chatbot')

openai_api_key = st.sidebar.text_input('OpenAI API Key')

def generate_response(input_text):
  llm = OpenAI(temperature=0.7, openai_api_key=openai_api_key)
  st.info(llm(input_text))

with st.form('my_form'):
    text = st.text_area(
        'Enter text:', 'What are the three key pieces of advice for learning how to code?'
    )
    submitted = st.form_submit_button('Submit')
    if not openai_api_key.startswith('sk-'):
        st.warning('Please enter your OpenAI API key!', icon='⚠')
    if submitted and openai_api_key.startswith('sk-'):
        generate_response(text)



from typing import Any
from pydantic import PrivateAttr
from langchain_community.tools.ddg_search.tool import (
    DuckDuckGoSearchResults,
    DuckDuckGoSearchAPIWrapper
)
from duckduckgo_search import DDGS
from langchain.tools import BaseTool


class DuckDuckGoSearchTool(BaseTool):
    """Web Search tool using Duck Duck Go API."""
    name: str = "DuckDuckGo Web Search"
    description: str = "Search the web using DuckDuckGo Search API"
    source: Any = None
    max_results: int = 5
    region: str = None

    def __init__(self, source: str = "news", results: int = 5, region: str = 'wt-wt', **kwargs: Any):
        super().__init__(**kwargs)
        self.source = source
        self.max_results = results
        self.region = region

    def _run(self, query: str) -> dict:
        """Run the DuckDuckGo Search Tool."""
        wrapper = DuckDuckGoSearchAPIWrapper(
            region=self.region,
            time="y",
            max_results=self.max_results
        )
        search = DuckDuckGoSearchResults(
            api_wrapper=wrapper,
            source=self.source
        )
        return search.run(query)

class DuckDuckGoRelevantSearch(BaseTool):
    """Web Search tool using Duck Duck Go API."""
    name: str = "DuckDuckGo Relevant Search"
    description: str = "Search the web using DuckDuckGo Search API"
    _max_results: PrivateAttr
    _region: PrivateAttr

    def __init__(self, results: int = 5, region: str = 'wt-wt', **kwargs: Any):
        super().__init__(**kwargs)
        self._max_results = results
        self._region = region

    def _run(
		self,
        query: str,
		**kwargs: Any,
	) -> Any:
        """Search Internet for relevant information based on a query."""
        search = DDGS()
        return search.text(
            keywords=query,
            region=self._region,
            safesearch='moderate',
            max_results=self._max_results
        )



"""
Data Management for Bots.
"""
from asyncdb.exceptions import NoDataFound, DriverError
from querysource.datasources.drivers.rethink import rethink_default
from navigator.views import FormModel
from ..utils.models import BotData
from ...interfaces import DBInterface

class DataManagement(DBInterface, FormModel):
    """
    Managing Data to be inserted into Vector Databases.
    """
    model = BotData
    path: str = '/api/v1/bot_management/data'

    async def get_rethink(self):
        params = rethink_default.params()
        return self.get_database('rethink', params=params)

    async def _get_form(self, args, qp, fields):
        """Get Form information."""
        bot = args.get('id', None)
        async with await self.get_rethink() as conn:
            await conn.use('navigator')
            filter = {}
            if bot:
                filter= {
                    "chatbot_id": bot
                }
                try:
                    result = await conn.fetch_all(
                        'chatbots_data',
                        filter=filter
                    )
                except NoDataFound:
                    return self.json_response(
                        status=204
                    )
            return self.json_response(
                result, status=200
            )

    async def put(self):
        """
        Creating a new data-store record.
        """
        data = await self.validate_payload()
        args = self.get_arguments()
        # unique field is ['chatbot_id', 'source_type', 'version']
        bot = args.get('id', data.chatbot_id)
        filter = {
            'chatbot_id': str(bot),
            'source_type': data.source_type,
            'version': data.version
        }
        tbl = data.Meta.name
        exists = False
        async with await self.get_rethink() as conn:
            await conn.use('navigator')
            try:
                result = await conn.fetch_one(
                    table=tbl,
                    filter=filter
                )
                if result:
                    exists = True
            except NoDataFound:
                exists = False
            if exists is False:
                try:
                    state = await conn.insert(
                        tbl, data.to_dict(), on_conflict='replace'
                    )
                    return self.json_response(
                        response={
                            "message": "Bot data inserted",
                            "state": state,
                        },
                        status=203
                    )
                except DriverError as exc:
                    return self.error(
                        response={
                            "error": "unable to insert RT data",
                            "exception": str(exc)
                        },
                        status=400
                    )
            else:
                data.version += data.version
                try:
                    state = await conn.update(
                        tbl, data.to_dict(), filter=filter, return_changes=True
                    )
                    return self.json_response(
                        response={
                            "message": "Bot data updated",
                            "state": state,
                        },
                        status=202
                    )
                except DriverError as exc:
                    return self.error(
                        response={
                            "error": "unable to insert RT data",
                            "exception": str(exc)
                        },
                        status=400
                    )

    async def patch(self):
        """
        Patching (updating) existing data.
        """
        # data = await self.validate_payload()
        data = await self.json_data()
        args = self.get_arguments()
        # unique field is ['chatbot_id', 'source_type', 'version']
        bot = args.get('id', data['chatbot_id'])
        try:
            filter = {
                'chatbot_id': str(bot),
                'source_type': data['source_type'],
                'version': data['version']
            }
        except KeyError:
            return self.error(
                response={
                    "error": "Invalid data for Data Filtering"
                },
                status=400
            )
        tbl = 'chatbots_data'
        async with await self.get_rethink() as conn:
            await conn.use('navigator')
            try:
                result = await conn.fetch_one(
                    table=tbl,
                    filter=filter
                )
                if result:
                    # update existing:
                    for k,v in data.items():
                        result[k] = v
                    try:
                        state = await conn.update(
                            tbl, result, filter=filter, return_changes=True
                        )
                        return self.json_response(
                            response={
                                "message": "Bot data updated",
                                "state": state,
                            },
                            status=202
                        )
                    except DriverError as exc:
                        return self.error(
                            response={
                                "error": "unable to Update RT data",
                                "exception": str(exc)
                            },
                            status=400
                        )
            except NoDataFound:
                return self.error(
                    response={
                        "error": f"Data not Found for {filter}",
                    },
                    status=400
                )

    async def delete(self):
        """
        Remove a data-store record
        """
        data = await self.validate_payload()
        args = self.get_arguments()
        # unique field is ['chatbot_id', 'source_type', 'version']
        bot = args.get('id', data.chatbot_id)
        filter = {
            'chatbot_id': str(bot),
            'source_type': data.source_type,
            'version': data.version
        }
        tbl = data.Meta.name
        async with await self.get_rethink() as conn:
            await conn.use('navigator')
            try:
                result = await conn.fetch_one(
                    table=tbl,
                    filter=filter
                )
                if result:
                    state = await conn.delete(
                        table=tbl,
                        filter=filter,
                        return_changes=True
                    )
                    return self.json_response(
                        response={
                            "message": "Bot data deleted",
                            "state": state,
                        },
                        status=202
                    )
            except NoDataFound:
                return self.error(
                    response={
                        "error": f"Data not Found for {filter}",
                    },
                    status=400
                )




from pathlib import Path, PurePath
from typing import Any
from collections.abc import Callable
import pandas as pd
from langchain.docstore.document import Document
from .abstract import AbstractLoader


class QAFileLoader(AbstractLoader):
    """
    Question and Answers File based on Excel.
    """
    _extension = ['.xlsx']
    chunk_size = 768

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'QA',
        columns: list = ['Question', 'Answer'],
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type, **kwargs)
        self.path = path
        self._columns = columns
        if isinstance(path, str):
            self.path = Path(path).resolve()
        if self.path.is_dir():
            raise ValueError(
                f"Currently only accepting single Files."
            )

    def _load_document(self, path: PurePath) -> list:
        if path.exists():
            print('Load QA Excel File: ', path)
            df = pd.read_excel(path)
            q = self._columns[0]
            a = self._columns[1]
            docs = []
            for idx, row in df.iterrows():
                # Question Document
                doc = Document(
                    page_content=f"**Question:** {row[q]}: **Answer:** {row[a]}",
                    metadata={
                        "url": '',
                        "index": f"{path.name} #{idx}",
                        "source": f"{path.name} Row.#{idx}",
                        "filename": f"{path.name}",
                        "question": row[q],
                        "answer": row[a],
                        "page_number": idx,
                        "source_type": self._source_type,
                        "type": "QA",
                        "summary": f"Question: {row[q]}?: **{row[a]}**",
                        "document_meta": {
                            "question": row[q],
                            "answer": row[a],
                        }
                    }
                )
                docs.append(doc)
            return docs
        return []

    def load(self, **kwargs) -> list:
        """
        Load Chapters from a PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self.path.is_file():
            documents = self._load_document(path=self.path, **kwargs)
            # after all documents are retrieved, procesed and stored
            return self.split_documents(documents)

    def parse(self, source):
        pass



from typing import Any
from collections.abc import Callable
from pathlib import Path, PurePath
import numpy as np
from PIL import Image
from langchain.docstore.document import Document
from transformers import CLIPModel
import torch
from torchvision import transforms
from .abstract import AbstractLoader
from ..stores.abstract import AbstractStore


class ImageLoader(AbstractLoader):
    """
    Image Loader.
    """
    _extension = ['.jpg', '.jpeg', '.png']
    chunk_size = 768

    def __init__(
        self,
        path: PurePath,
        store: AbstractStore,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'image',
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type, **kwargs)
        self.path = path
        if isinstance(path, str):
            self.path = Path(path).resolve()
        # Model:
        self._model = CLIPModel.from_pretrained(
            # "openai/clip-vit-base-patch32"
            "openai/clip-vit-large-patch14-336"
        )
        # Define image preprocessing
        self._preprocess = transforms.Compose(
            [
                transforms.Resize((336, 336)),  # Adjust the size to match the model's expected input
                transforms.CenterCrop(336),  # Optionally add a center crop if needed
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)
                )  # CLIP's original normalization
            ]
        )
        # required Milvus Store:
        self.store = store

    def transform_image(self, img_data):
        image = self._preprocess(img_data)
        image = image.unsqueeze(0)
        with torch.no_grad():
            features = self._model.get_image_features(pixel_values=image)
        embedding = features.squeeze().cpu().numpy()
        return embedding.astype(np.float32)

    def _insert_image(self, data):
        return self.store.insert(data)

    def _load_image(self, path) -> list:
        """
        Load an Image file.
        Args:
            path (Path): The path to the Image file.
        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading Image file: {path}")
            img = Image.open(path).convert('RGB')
            embedding = self.transform_image(img).tolist()
            data={
                "url": '',
                "source": f"{path.name}",
                "filename": path,
                "question": '',
                "answer": '',
                "source_type": self._source_type,
                "type": "image",
                "text": '',
                "vector": embedding,
                "document_meta": {
                    "image": path.name,
                    "extension": path.suffix
                }
            }
            self._insert_image([embedding])
        return []

    def load(self) -> list:
        """
        Load data from a Image file.
        Returns:
            list: A list of Langchain Documents.
        """
        if not self.path.exists():
            raise FileNotFoundError(f"Image file/directory not found: {self.path}")
        if self.path.is_dir():
            # iterate over the files in the directory
            for ext in self._extension:
                for item in self.path.glob(f'*{ext}'):
                    self._load_image(item)
        elif self.path.is_file():
            self._load_image(self.path)
        else:
            raise ValueError(
                f"Image Loader: Invalid path: {self.path}"
            )
        # Load Image loads the image directly to database.
        return True

    def parse(self, source):
        raise NotImplementedError(
            "Parser method is not implemented for ImageLoader."
        )



"""
Settings Configuration for Parrot-AI.

Basic Configuration.
"""
from navconfig import config, BASE_DIR
from navconfig.logging import logging


logging.getLogger(name='sentence_transformers').setLevel(logging.WARNING)

AUTHENTICATION_BACKENDS = (
    'navigator_auth.backends.AzureAuth',
    'navigator_auth.backends.APIKeyAuth',
    'navigator_auth.backends.TokenAuth',
    'navigator_auth.backends.TrocToken',
    'navigator_auth.backends.ADFSAuth',
    'navigator_auth.backends.Oauth2Provider',
    'navigator_auth.backends.BasicAuth'
)

# Azure Bot:
MS_TENANT_ID = config.get('MS_TENANT_ID')
MS_CLIENT_ID = config.get('MS_CLIENT_ID')
MS_CLIENT_SECRET = config.get('MS_CLIENT_SECRET')

# New Bot:
NEW_CLIENT_ID = config.get('NEW_CLIENT_ID')
NEW_CLIENT_SECRET = config.get('NEW_CLIENT_SECRET')

# Other Bots:
BOSE_CLIENT_ID = config.get('BOSE_CLIENT_ID')
BOSE_CLIENT_SECRET = config.get('BOSE_CLIENT_SECRET')
ODOO_CLIENT_ID = config.get('ODOO_CLIENT_ID')
ODOO_CLIENT_SECRET = config.get('ODOO_CLIENT_SECRET')
ASKBRETT_CLIENT_ID = config.get('ASKBRETT_CLIENT_ID')
ASKBRETT_CLIENT_SECRET = config.get('ASKBRETT_CLIENT_SECRET')
TROCERS_CLIENT_ID = config.get('TROCERS_CLIENT_ID')
TROCERS_CLIENT_SECRET = config.get('TROCERS_CLIENT_SECRET')
BOTTROCDEV_CLIENT_ID = config.get('BOTTROCDEV_CLIENT_ID')
BOTTROCDEV_CLIENT_SECRET = config.get('BOTTROCDEV_CLIENT_SECRET')
ATTBOT_CLIENT_ID = config.get('ATTBOT_CLIENT_ID')
ATTBOT_CLIENT_SECRET = config.get('ATTBOT_CLIENT_SECRET')


# ScyllaDB Configuration:
SCYLLADB_HOST = config.get('SCYLLADB_HOST', fallback='localhost')
SCYLLADB_PORT = int(config.get('SCYLLADB_PORT', fallback=9042))
SCYLLADB_USERNAME = config.get('SCYLLADB_USERNAME', fallback='navigator')
SCYLLADB_PASSWORD = config.get('SCYLLADB_PASSWORD', fallback='navigator')
SCYLLADB_KEYSPACE = config.get('SCYLLADB_KEYSPACE', fallback='navigator')



from pathlib import PurePath
from langchain_community.document_loaders.parsers import LanguageParser
from langchain_community.document_loaders import GithubFileLoader
from langchain_text_splitters import Language
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter
)
from navconfig import config
from .abstract import AbstractLoader

class GithubLoader(AbstractLoader):
    """
    Load Code from a Github Repository.
    """
    def __init__(
        self,
        repository_url: str,
        lang: str = 'python',
        source_type: str = 'code',
        branch: str = 'main',
        **kwargs
    ):
        super().__init__(source_type=source_type, **kwargs)
        self._url = repository_url
        self.branch = branch
        self.github_token = kwargs.get('github_token', config.get('GITHUB_TOKEN'))
        self.lang = lang
        if lang == 'python':
            self.parser = LanguageParser(language=Language.PYTHON, parser_threshold=500)
            self.splitter = RecursiveCharacterTextSplitter.from_language(
                language=Language.PYTHON, chunk_size=1024, chunk_overlap=200
            )
            self.suffixes = [".py", ".pyx"]
        elif lang == 'javascript':
            self.parser = LanguageParser(language=Language.JS, parser_threshold=500)
            self.splitter = RecursiveCharacterTextSplitter.from_language(
                language=Language.JS, chunk_size=1024, chunk_overlap=200
            )
            self.suffixes = [".js", ".jsx", ".json", ".ts", ".tsx"]
        elif lang == 'typescript':
            self.parser = LanguageParser(language=Language.TS, parser_threshold=500)
            self.splitter = RecursiveCharacterTextSplitter.from_language(
                language=Language.TS, chunk_size=1024, chunk_overlap=200
            )
            self.suffixes = [".js", ".jsx", ".json", ".ts", ".tsx"]
        else:
            raise ValueError(
                f"Language {lang} not supported for Repository"
            )

    def load(self) -> list:
        self.logger.info(f'Loading Github Repository > {self._url}:<{self.branch}>')
        loader = GithubFileLoader(
            repo=self._url,
            github_api_url='https://api.github.com',
            access_token=self.github_token,
            branch=self.branch,
        )
        docs = loader.load()
        for doc in docs:
            doc.metadata['source_type'] = self._source_type
        return self.split_documents(docs)

    def parse(self, source):
        raise NotImplementedError("Parser method is not implemented for PDFLoader.")



from pathlib import PurePath
from langchain_community.document_loaders import (
    UnstructuredPowerPointLoader
)
from .abstract import AbstractLoader


class PPTXLoader(AbstractLoader):
    """
    Loader for PPTX files.
    """
    _extension: list = ['.pptx']

    def load(self, path: PurePath) -> list:
        if self._check_path(path):
            docs = []
            self.logger.info(f"Loading PPTX file: {path}")
            ppt_loader = UnstructuredPowerPointLoader(
                file_path=str(path)
            )
            docs += ppt_loader.load()
            for doc in docs:
                doc.metadata['source_type'] = self._source_type
            # Split the documents into chunks
            return self.split_documents(docs)
        else:
            return []

    def parse(self, source):
        pass



"""
Foundational base of every Chatbot and Agent in ai-parrot.
"""
from abc import ABC
from collections.abc import Callable
from typing import Any, Union
from pathlib import Path, PurePath
import uuid
from aiohttp import web
import torch
from transformers import (
    AutoModel,
    AutoConfig,
    AutoTokenizer,
)
# Langchain
from langchain.docstore.document import Document
from langchain.memory import (
    ConversationBufferMemory
)
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter
)
from langchain_community.chat_message_histories import (
    RedisChatMessageHistory
)
# Navconfig
from navconfig import BASE_DIR
from navconfig.exceptions import ConfigError  # pylint: disable=E0611
from navconfig.logging import logging
from asyncdb.exceptions import NoDataFound

## LLM configuration
from ..llms import get_llm, AbstractLLM

## Vector Database configuration:
from ..stores import get_vectordb

from ..utils import SafeDict, parse_toml_config
from .retrievals import RetrievalManager
from ..conf import (
    EMBEDDING_DEVICE,
    MAX_VRAM_AVAILABLE,
    RAM_AVAILABLE,
    default_dsn,
    REDIS_HISTORY_URL,
    EMBEDDING_DEFAULT_MODEL
)
from ..interfaces import DBInterface
from ..models import ChatbotModel


class AbstractChatbot(ABC, DBInterface):
    """Represents an Chatbot in Navigator.

        Each Chatbot has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """

    template_prompt: str = (
        "You are {name}, an expert AI assistant and {role} Working at {company}.\n\n"
        "Your primary function is to {goal}\n"
        "Use the provided context of the documents you have processed or extracted from other provided tools or sources to provide informative, detailed and accurate responses.\n"
        "I am here to help with {role}.\n"
        "**Backstory:**\n"
        "{backstory}.\n\n"
        "Focus on answering the question directly but detailed. Do not include an introduction or greeting in your response.\n\n"
        "{company_information}\n\n"
        "Here is a brief summary of relevant information:\n"
        "Context: {context}\n\n"
        "Given this information, please provide answers to the following question adding detailed and useful insights:\n\n"
        "**Chat History:** {chat_history}\n\n"
        "**Human Question:** {question}\n"
        "Assistant Answer:\n\n"
        "{rationale}\n"
        "You are a fluent speaker, you can talk and respond fluently in English and Spanish, and you must answer in the same language as the user's question. If the user's language is not English, you should translate your response into their language.\n"
    )

    def _get_default_attr(self, key, default: Any = None, **kwargs):
        if key in kwargs:
            return kwargs.get(key)
        if hasattr(self, key):
            return getattr(self, key)
        if not hasattr(self, key):
            return default
        return getattr(self, key)

    def __init__(self, **kwargs):
        """Initialize the Chatbot with the given configuration."""
        # Start initialization:
        self.kb = None
        self.knowledge_base: list = []
        # Chatbot ID:
        self.chatbot_id: uuid.UUID = kwargs.get(
            'chatbot_id',
            str(uuid.uuid4().hex)
        )
        # Basic Information:
        self.name = self._get_default_attr(
            'name', 'NAV', **kwargs
        )
        ##  Logging:
        self.logger = logging.getLogger(f'{self.name}.Chatbot')
        self.description = self._get_default_attr(
            'description', 'Navigator Chatbot', **kwargs
        )
        self.role = self._get_default_attr(
            'role', 'Chatbot', **kwargs
        )
        self.goal = self._get_default_attr(
            'goal',
            'provide helpful information to users',
            **kwargs
        )
        self.backstory = self._get_default_attr(
            'backstory',
            default=self.default_backstory(),
            **kwargs
        )
        self.rationale = self._get_default_attr(
            'rationale',
            default=self.default_rationale(),
            **kwargs
        )
        # Configuration File:
        self.config_file: PurePath = kwargs.get('config_file', None)
        # Other Configuration
        self.confidence_threshold: float = kwargs.get('threshold', 0.5)
        self.context = kwargs.pop('context', '')

        # Company Information:
        self.company_information: dict = kwargs.pop('company_information', {})

        # Pre-Instructions:
        self.pre_instructions: list = kwargs.get(
            'pre_instructions',
            []
        )

        # Knowledge base:
        self.knowledge_base: list = []
        self._documents_: list = []

        # Text Documents
        self.documents_dir = kwargs.get(
            'documents_dir',
            None
        )
        if isinstance(self.documents_dir, str):
            self.documents_dir = Path(self.documents_dir)
        if not self.documents_dir:
            self.documents_dir = BASE_DIR.joinpath('documents')
        if not self.documents_dir.exists():
            self.documents_dir.mkdir(
                parents=True,
                exist_ok=True
            )
        # Models, Embed and collections
        # Vector information:
        self.chunk_size: int = int(kwargs.get('chunk_size', 768))
        self.dimension: int = int(kwargs.get('dimension', 768))
        self._database: dict = kwargs.get('database', {})
        self._store: Callable = None
        # Embedding Model Name
        self.use_bge: bool = bool(
            kwargs.get('use_bge', 'False')
        )
        self.use_fastembed: bool = bool(
            kwargs.get('use_fastembed', 'False')
        )
        self.embedding_model_name = kwargs.get(
            'embedding_model_name', None
        )
        # embedding object:
        self.embeddings = kwargs.get('embeddings', None)
        self.tokenizer_model_name = kwargs.get(
            'tokenizer', None
        )
        self.summarization_model = kwargs.get(
            'summarization_model',
            "facebook/bart-large-cnn"
        )
        self.rag_model = kwargs.get(
            'rag_model',
            "rlm/rag-prompt-llama"
        )
        self._text_splitter_model = kwargs.get(
            'text_splitter',
            'mixedbread-ai/mxbai-embed-large-v1'
        )
        # Definition of LLM
        # Overrriding LLM object
        self._llm_obj: Callable = kwargs.get('llm', None)
        # LLM base Object:
        self._llm: Callable = None

        # Max VRAM usage:
        self._max_vram = int(
            kwargs.get('max_vram', MAX_VRAM_AVAILABLE)
        )

    def get_llm(self):
        return self._llm_obj

    def __repr__(self):
        return f"<Chatbot.{self.__class__.__name__}:{self.name}>"

    # Database:
    @property
    def store(self):
        if not self._store.connected:
            self._store.connect()
        return self._store

    def default_rationale(self) -> str:
        # TODO: read rationale from a file
        return (
            "I am a language model trained by Google.\n"
            "I am designed to provide helpful information to users."
            "Remember to maintain a professional tone."
            "If I cannot find relevant information in the documents,"
            "I will indicate this and suggest alternative avenues for the user to find an answer."
        )

    def default_backstory(self) -> str:
        return (
            "help with Human Resources related queries or knowledge-based questions about T-ROC Global.\n"
            "You can ask me about the company's products and services, the company's culture, the company's clients.\n"
            "You have the capability to read and understand various Human Resources documents, "
            "such as employee handbooks, policy documents, onboarding materials, company's website, and more.\n"
            "I can also provide information about the company's policies and procedures, benefits, and other HR-related topics."
        )

    async def configure(self, app = None) -> None:
        if app is None:
            self.app = None
        else:
            if isinstance(app, web.Application):
                self.app = app  # register the app into the Extension
            else:
                self.app = app.get_app()  # Nav Application
        # Config File:
        config_file = BASE_DIR.joinpath(
            'etc',
            'config',
            'chatbots',
            self.name.lower(),
            "config.toml"
        )
        if config_file.exists():
            self.logger.notice(
                f"Loading Bot {self.name} from config: {config_file.name}"
            )
        if (bot := await self.bot_exists(name=self.name, uuid=self.chatbot_id)):
            self.logger.notice(
                f"Loading Bot {self.name} from Database: {bot.chatbot_id}"
            )
            # Bot exists on Database, Configure from the Database
            await self.from_database(bot, config_file)
        elif config_file.exists():
            # Configure from the TOML file
            await self.from_config_file(config_file)
        else:
            raise ValueError(
                f'Bad configuration procedure for bot {self.name}'
            )
        # adding this configured chatbot to app:
        if self.app:
            self.app[f"{self.name.lower()}_chatbot"] = self

    def _configure_llm(self, llm, config):
        """
        Configuration of LLM.
        """
        if isinstance(self._llm_obj, AbstractLLM):
            self._llm = self._llm_obj.get_llm()
        elif self._llm_obj is not None:
            self._llm = self._llm_obj
        else:
            if llm:
                # LLM:
                self._llm_obj = get_llm(
                    llm,
                    **config
                )
                # getting langchain LLM from Obj:
                self._llm = self._llm_obj.get_llm()
            else:
                raise ValueError(
                    f"{self.name}: LLM is not defined in bot Configuration."
                )

    def _from_bot(self, bot, key, config, default) -> Any:
        value = getattr(bot, key, None)
        file_value = config.get(key, default)
        return value if value else file_value

    def _from_db(self, botobj, key, default = None) -> Any:
        value = getattr(botobj, key, default)
        return value if value else default

    async def bot_exists(
        self,
        name: str = None,
        uuid: uuid.UUID = None
    ) -> Union[ChatbotModel, bool]:
        """Check if the Chatbot exists in the Database."""
        db = self.get_database('pg', dsn=default_dsn)
        async with await db.connection() as conn:  # pylint: disable=E1101
            ChatbotModel.Meta.connection = conn
            try:
                if self.chatbot_id:
                    try:
                        bot = await ChatbotModel.get(chatbot_id=uuid)
                    except Exception:
                        bot = await ChatbotModel.get(name=name)
                else:
                    bot = await ChatbotModel.get(name=self.name)
                if bot:
                    return bot
            except NoDataFound:
                return False

    async def from_database(
        self,
        bot: Union[ChatbotModel, None] = None,
        config_file: PurePath = None
    ) -> None:
        """Load the Chatbot Configuration from the Database."""
        if not bot:
            db = self.get_database('pg', dsn=default_dsn)
            async with await db.connection() as conn:  # pylint: disable=E1101
                # import model
                ChatbotModel.Meta.connection = conn
                try:
                    if self.chatbot_id:
                        try:
                            bot = await ChatbotModel.get(chatbot_id=self.chatbot_id)
                        except Exception:
                            bot = await ChatbotModel.get(name=self.name)
                    else:
                        bot = await ChatbotModel.get(name=self.name)
                except NoDataFound:
                    # Fallback to File configuration:
                    raise ConfigError(
                        f"Chatbot {self.name} not found in the database."
                    )
        # Start Bot configuration from Database:
        if config_file and config_file.exists():
            file_config = await parse_toml_config(config_file)
            # Knowledge Base come from file:
            # Contextual knowledge-base
            self.kb = file_config.get('knowledge-base', [])
            if self.kb:
                self.knowledge_base = self.create_kb(
                    self.kb.get('data', [])
                )
        self.name = self._from_db(bot, 'name', default=self.name)
        self.chatbot_id = str(self._from_db(bot, 'chatbot_id', default=self.chatbot_id))
        self.description = self._from_db(bot, 'description', default=self.description)
        self.role = self._from_db(bot, 'role', default=self.role)
        self.goal = self._from_db(bot, 'goal', default=self.goal)
        self.rationale = self._from_db(bot, 'rationale', default=self.rationale)
        self.backstory = self._from_db(bot, 'backstory', default=self.backstory)
        # company information:
        self.company_information = self._from_db(
            bot, 'company_information', default=self.company_information
        )
        # LLM Configuration:
        llm = self._from_db(bot, 'llm', default='VertexLLM')
        llm_config = self._from_db(bot, 'llm_config', default={})
        # Configuration of LLM:
        self._configure_llm(llm, llm_config)
        # Other models:
        self.embedding_model_name = self._from_db(
            bot, 'embedding_name', None
        )
        self.tokenizer_model_name = self._from_db(
            bot, 'tokenizer', None
        )
        self.summarization_model = self._from_db(
            bot, 'summarize_model', "facebook/bart-large-cnn"
        )
        self.classification_model = self._from_db(
            bot, 'classification_model', None
        )
        # Database Configuration:
        db_config = bot.database
        vector_db = db_config.pop('vector_database')
        await self.store_configuration(vector_db, db_config)
        # after configuration, setup the chatbot
        if bot.template_prompt:
            self.template_prompt = bot.template_prompt
        self._define_prompt(
            config={}
        )

    async def from_config_file(self, config_file: PurePath) -> None:
        """Load the Chatbot Configuration from the TOML file."""
        self.logger.debug(
            f"Using Config File: {config_file}"
        )
        file_config = await parse_toml_config(config_file)
        # getting the configuration from config
        self.config_file = config_file
        # basic config
        basic = file_config.get('chatbot', {})
        # Chatbot Name:
        self.name = basic.get('name', self.name)
        self.description = basic.get('description', self.description)
        self.role = basic.get('role', self.role)
        self.goal = basic.get('goal', self.goal)
        self.rationale = basic.get('rationale', self.rationale)
        self.backstory = basic.get('backstory', self.backstory)
        # Company Information:
        self.company_information = basic.get(
            'company_information',
            self.company_information
        )
        # Model Information:
        llminfo = file_config.get('llm')
        llm = llminfo.get('llm', 'VertexLLM')
        cfg = llminfo.get('config', {})
        # Configuration of LLM:
        self._configure_llm(llm, cfg)

        # Other models:
        models = file_config.get('models', {})
        if not self.embedding_model_name:
            self.embedding_model_name = models.get(
                'embedding', EMBEDDING_DEFAULT_MODEL
            )
        if not self.tokenizer_model_name:
            self.tokenizer_model_name = models.get('tokenizer')
        if not self.embedding_model_name:
            # Getting the Embedding Model from the LLM
            self.embeddings = self._llm_obj.get_embedding()
        self.use_bge = models.get('use_bge', False)
        self.use_fastembed = models.get('use_fastembed', False)
        self.summarization_model = models.get(
            'summarize_model',
            "facebook/bart-large-cnn"
        )
        self.classification_model = models.get(
            'classification_model',
            None
        )
        # pre-instructions
        instructions = file_config.get('pre-instructions')
        if instructions:
            self.pre_instructions = instructions.get('instructions', [])
        # Contextual knowledge-base
        self.kb = file_config.get('knowledge-base', [])
        if self.kb:
            self.knowledge_base = self.create_kb(
                self.kb.get('data', [])
            )
        vector_config = file_config.get('database', {})
        vector_db = vector_config.pop('vector_database')
        # configure vector database:
        await self.store_configuration(
            vector_db,
            vector_config
        )
        # after configuration, setup the chatbot
        if 'template_prompt' in basic:
            self.template_prompt = basic.get('template_prompt')
        self._define_prompt(
            config=basic
        )

    def create_kb(self, documents: list):
        new_docs = []
        for doc in documents:
            content = doc.pop('content')
            source = doc.pop('source', 'knowledge-base')
            if doc:
                meta = {
                    'source': source,
                    **doc
                }
            else:
                meta = { 'source': source}
            if content:
                new_docs.append(
                    Document(
                        page_content=content,
                        metadata=meta
                    )
                )
        return new_docs

    async def store_configuration(self, vector_db: str, config: dict):
        """Create the Vector Store Configuration."""
        self.collection_name = config.get('collection_name')
        if not self.embeddings:
            embed = self.embedding_model_name
        else:
            embed = self.embeddings
        # TODO: add dynamic configuration of VectorStore
        self._store = get_vectordb(
            vector_db,
            embeddings=embed,
            use_bge=self.use_bge,
            use_fastembed=self.use_fastembed,
            **config
        )

    def _define_prompt(self, config: dict):
        # setup the prompt variables:
        for key, val in config.items():
            setattr(self, key, val)
        if self.company_information:
            self.template_prompt = self.template_prompt.format_map(
                SafeDict(
                    company_information=(
                        "For further inquiries or detailed information, you can contact us at:\n"
                        "- Contact Information: {contact_email}\n"
                        "- Use our contact form: {contact_form}\n"
                        "- or Visit our website: {company_website}\n"
                    )
                )
            )
        # Parsing the Template:
        self.template_prompt = self.template_prompt.format_map(
            SafeDict(
                name=self.name,
                role=self.role,
                goal=self.goal,
                backstory=self.backstory,
                rationale=self.rationale,
                threshold=self.confidence_threshold,
                **self.company_information
            )
        )
        # print('Template Prompt:', self.template_prompt)

    @property
    def llm(self):
        return self._llm

    @llm.setter
    def llm(self, model):
        self._llm_obj = model
        self._llm = model.get_llm()

    def _get_device(self, cuda_number: int = 0):
        torch.backends.cudnn.deterministic = True
        if torch.cuda.is_available():
            # Use CUDA GPU if available
            device = torch.device(f'cuda:{cuda_number}')
        elif torch.backends.mps.is_available():
            # Use CUDA Multi-Processing Service if available
            device = torch.device("mps")
        elif EMBEDDING_DEVICE == 'cuda':
            device = torch.device(f'cuda:{cuda_number}')
        else:
            device = torch.device(EMBEDDING_DEVICE)
        return device

    def get_tokenizer(self, model_name: str, chunk_size: int = 768):
        return AutoTokenizer.from_pretrained(
            model_name,
            chunk_size=chunk_size
        )

    def get_model(self, model_name: str):
        device = self._get_device()
        self._model_config = AutoConfig.from_pretrained(
            model_name, trust_remote_code=True
        )
        return AutoModel.from_pretrained(
            model_name,
            trust_remote_code=True,
            config=self._model_config,
            unpad_inputs=True,
            use_memory_efficient_attention=True,
        ).to(device)

    def get_text_splitter(self, model, chunk_size: int = 1024, overlap: int = 100):
        return RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
            model,
            chunk_size=chunk_size,
            chunk_overlap=overlap,
            add_start_index=True,  # If `True`, includes chunk's start index in metadata
            strip_whitespace=True,  # strips whitespace from the start and end
            separators=["\n\n", "\n", "\r\n", "\r", "\f", "\v", "\x0b", "\x0c"],
        )

    def chunk_documents(self, documents, chunk_size):
        # Yield successive n-sized chunks from documents.
        for i in range(0, len(documents), chunk_size):
            yield documents[i:i + chunk_size]

    def get_available_vram(self):
        """
        Returns available VRAM in megabytes.
        """
        try:
            # Clear any unused memory to get a fresher estimate
            torch.cuda.empty_cache()
            # Convert to MB
            total_memory = torch.cuda.get_device_properties(0).total_memory / (1024 ** 2)
            reserved_memory = torch.cuda.memory_reserved(0) / (1024 ** 2)
            available_memory = total_memory - reserved_memory
            self.logger.notice(f'Available VRAM : {available_memory}')
            # Limit by predefined max usage
            return min(available_memory, self._max_vram)
        except RuntimeError:
            # Limit by predefined max usage
            return min(RAM_AVAILABLE, self._max_vram)

    def _estimate_chunk_size(self):
        """Estimate chunk size based on VRAM usage.
        This is a simplistic heuristic and might need tuning based on empirical data
        """
        available_vram = self.get_available_vram()
        estimated_vram_per_doc = 50  # Estimated VRAM in megabytes per document, adjust based on empirical observation
        chunk_size = max(1, int(available_vram / estimated_vram_per_doc))
        self.logger.notice(
            f'Chunk size for Load Documents: {chunk_size}'
        )
        return chunk_size

    ## Utility Loaders
    ##

    async def load_documents(
        self,
        documents: list,
        collection: str = None,
        delete: bool = False
    ):
        # Load Raw Documents into the Vectorstore
        print('::: LEN >> ', len(documents), type(documents))
        if len(documents) < 1:
            self.logger.warning(
                "There is no documents to be loaded, skipping."
            )
            return

        self._documents_.extend(documents)
        if not collection:
            collection = self.collection_name

        self.logger.notice(f'Loading Documents: {len(documents)}')
        document_chunks = self.chunk_documents(
            documents,
            self._estimate_chunk_size()
        )
        async with self._store.connection(alias='default') as store:
            # if delete is True, then delete the collection
            if delete is True:
                await store.delete_collection(collection)
                fdoc = documents.pop(0)
                await store.create_collection(
                    collection,
                    fdoc
                )
            for chunk in document_chunks:
                await store.load_documents(
                    chunk,
                    collection=collection
                )

    def clean_history(
        self,
        session_id: str = None
    ):
        try:
            redis_client = RedisChatMessageHistory(
                url=REDIS_HISTORY_URL,
                session_id=session_id,
                ttl=60
            )
            redis_client.clear()
        except Exception as e:
            self.logger.error(
                f"Error clearing chat history: {e}"
            )

    def get_memory(
        self,
        session_id: str = None,
        key: str = 'chat_history',
        input_key: str = 'question',
        output_key: str = 'answer',
        size: int = 5,
        ttl: int = 86400
    ):
        args = {
            'memory_key': key,
            'input_key': input_key,
            'output_key': output_key,
            'return_messages': True,
            'max_len': size
        }
        if session_id:
            message_history = RedisChatMessageHistory(
                url=REDIS_HISTORY_URL,
                session_id=session_id,
                ttl=ttl
            )
            args['chat_memory'] = message_history
        return ConversationBufferMemory(
            **args
        )

    def get_retrieval(self, source_path: str = 'web', request: web.Request = None):
        pre_context = "\n".join(f"- {a}." for a in self.pre_instructions)
        custom_template = self.template_prompt.format_map(
            SafeDict(
                summaries=pre_context
            )
        )
        # Generate the Retrieval
        rm = RetrievalManager(
            chatbot_id=self.chatbot_id,
            chatbot_name=self.name,
            source_path=source_path,
            model=self._llm,
            store=self._store,
            memory=None,
            template=custom_template,
            kb=self.knowledge_base,
            request=request
        )
        return rm



from collections.abc import Callable
from pathlib import PurePath
from typing import Union
from io import StringIO
import mimetypes
import magic
import xlrd
import numpy as np
import pandas
from pandas._libs.parsers import STR_NA_VALUES  # pylint: disable=E0611
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter
)
from langchain.docstore.document import Document
from navigator.libs.json import JSONContent
from .abstract import AbstractLoader


excel_based = (
    "application/vnd.ms-excel.sheet.binary.macroEnabled.12",
    "application/vnd.ms-excel.sheet.macroEnabled.12",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "application/vnd.ms-excel"
)

supported_extensions = (
    ".xls",
    ".xlsx",
    ".xlsm",
    ".xlsb"
)

class ExcelLoader(AbstractLoader):
    """
    Loader for Excel files using Pandas (preserving Table structure).
    """

    _extension = supported_extensions

    def __init__(
        self,
        path: Union[list[PurePath], PurePath],
        tokenizer: Callable = None,
        text_splitter: Callable = None,
        source_type: str = 'Excel',
        **kwargs
    ):
        self.path = path
        # Index Key:
        self._index_keys: list = kwargs.pop('index', [])
        self._type: str = kwargs.pop('document_type', 'Excel')
        self.sheet_name = kwargs.pop('sheet_name', "Sheet 1")
        self.skiprows = kwargs.pop('skiprows', 0)
        self._pdargs = kwargs.pop('pd_args', {})
        self._magic = magic.Magic(mime=True)
        self.mimetype = kwargs.pop('mimetype', None)
        self.filter_nan: bool = kwargs.pop('filter_nan', True)
        self.na_values: list = ["NULL", "TBD"]
        # Operations over dataframe:
        self._drop_empty = kwargs.pop('drop_empty', False)
        self._trim = kwargs.pop('trim', False)
        self._infer_types = kwargs.pop('infer_types', True)
        self._fillna: bool = kwargs.pop('fillna', False)
        self._rename_cols = kwargs.pop('rename_cols', {})
        self._to_integer: list = kwargs.pop('to_integer', [])
        super().__init__(
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            **kwargs
        )
        # JSON encoder:
        self._encoder = JSONContent()

    def clean_empty(self, df: pandas.DataFrame, columns: list = None) -> pandas.DataFrame:
        """
        Clean empty rows and columns from a DataFrame.

        Args:
            df (pd.DataFrame): The DataFrame to clean.

        Returns:
            pd.DataFrame: The cleaned DataFrame.
        """
        df.dropna(axis=1, how="all", inplace=True)
        df.dropna(axis=0, how="all", inplace=True)
        if columns:
            for column in columns:
                condition = df[
                    (df[column].empty)
                    | (df[column] == "")
                    | (df[column].isna())
                ].index
                df.drop(condition, inplace=True)
        return df

    def open_excel(self, filename: PurePath) -> pandas.DataFrame:
        self.logger.debug(
            f"Opening Excel file {filename}"
        )
        ## Define NA Values:
        default_missing = STR_NA_VALUES.copy()
        for val in self.na_values:  # pylint: disable=E0203
            default_missing.add(val)
            default_missing.add(val)
        self.na_values = default_missing
        if filename.suffix not in supported_extensions:
            raise ValueError(
                f"Unsupported Excel file format: {filename.suffix}"
            )
        if not filename.exists():
            raise FileNotFoundError(
                f"Excel file not found: {filename}"
            )
        if not self.mimetype:
            try:
                self.mimetype = self._magic.from_file(str(filename))
                self.logger.debug(f":: Detected MIME IS: {self.mimetype}")
            except Exception as exc:
                self.logger.error(f":: Error detecting MIME: {exc}")
                self.mimetype = mimetypes.guess_type(str(filename))[0]
            if not self.mimetype:
                # Cannot Detect Mime type:
                ext = filename.suffix
                if ext == ".xlsx" or ext == ".xls":
                    self.mimetype = "application/vnd.ms-excel"
                elif ext == ".csv" or ext == ".txt":
                    self.mimetype = "text/csv"
        if (
            self.mimetype == "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        ):
            # xlsx or any openxml based document
            file_engine = "openpyxl"
        elif self.mimetype == "application/vnd.ms-excel.sheet.binary.macroEnabled.12":
            # xlsb
            file_engine = "pyxlsb"
        else:
            # Using extension:
            ext = filename.suffix
            if ext == ".xlsx":
                file_engine = "openpyxl"
            elif ext == ".xls":
                file_engine = "xlrd"
            elif ext == ".xlsb":
                file_engine = "pyxlsb"
            else:
                raise ValueError(
                    f"Unsupported Excel file format: {filename.suffix}"
                )
        # Build Arguments:
        pd_args = self._pdargs.copy()
        pd_args['sheet_name'] = self.sheet_name
        pd_args['skiprows'] = self.skiprows
        pd_args['engine'] = file_engine
        pd_args['na_values'] = self.na_values
        pd_args['na_filter'] = self.filter_nan
        try:
            df = pandas.read_excel(
                filename,
                keep_default_na=False,
                **pd_args
            )
        except (IndexError, xlrd.biffh.XLRDError) as err:
            raise ValueError(
                f"Excel Index error on File {filename}: {err}"
            ) from err
        except pandas.errors.EmptyDataError as err:
            raise ValueError(f"Empty File {filename}: {err}") from err
        except pandas.errors.ParserError as err:
            raise ValueError(f"Error Parsing File {filename}: {err}") from err
        except Exception as err:
            self.logger.exception(str(err), stack_info=True)
            raise
        # Post-Processing the DataFrame:
        if self._infer_types is True:
            df.infer_objects()
        # rename cols:
        if self._rename_cols:
            try:
                # Renaming Pandas Columns:
                df.rename(columns=self._rename_cols, inplace=True)
            except Exception as err:
                self.logger.error(
                    f"Error Renaming Columns: {err}"
                )
        # Clean Empty Rows:
        df = self.clean_empty(df, self._index_keys)
        if self._drop_empty:
            df.dropna(axis=1, how="all", inplace=True)
            df.dropna(axis=0, how="all", inplace=True)
            df = df.loc[:, ~df.columns.str.contains("^Unnamed")]
        if self._trim:
            cols = df.select_dtypes(include=["object", "string"])
            for col in cols:
                df[col] = df[col].astype(str).str.strip()
            # Trim whitespace from all string columns
            # df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
        if self._to_integer:
            for column in self._to_integer:
                df[column] = df[column].fillna(0)
                try:
                    df[column] = pandas.to_numeric(df[column], errors="coerce")
                    df[column] = df[column].astype('Int64')
                except Exception:
                    continue
        if self._fillna:
            # Select all integer columns (both nullable and non-nullable)
            int_columns = df.select_dtypes(
                include=['int64', 'Int64', 'int32', 'Int32']
            ).columns  # Add other integer types if necessary
            # Fill NaN values with zeros in those columns
            df[int_columns] = df[int_columns].fillna(0)
            # Select the Strings:
            str_columns = df.select_dtypes(include=["object", "string"]).columns
            df[str_columns] = df[str_columns].astype(str).replace(["nan", np.nan], "", regex=True)
        print(df)
        print("::: Printing Column Information === ")
        for column, t in df.dtypes.items():
            print(column, "->", t, "->", df[column].iloc[0])
        return df

    def unique_columns(self, df: pandas.DataFrame) -> pandas.DataFrame:
        """
        Rename duplicate columns in the DataFrame to ensure they are unique.

        Args:
            df (pd.DataFrame): The DataFrame with potential duplicate column names.

        Returns:
            pd.DataFrame: A DataFrame with unique column names.
        """
        seen = {}
        new_columns = []
        for col in df.columns:
            new_col = col
            count = seen.get(col, 0)
            while new_col in new_columns:
                count += 1
                new_col = f"{col}_{count}"
            new_columns.append(new_col)
            seen[col] = count
        df.columns = new_columns
        return df

    def get_json(self, df: pandas.DataFrame) -> str:
        """
        Convert a DataFrame to a JSON string.

        Args:
            df (pd.DataFrame): The DataFrame to convert.

        Returns:
            str: The JSON string.
        """
        buffer = StringIO()
        df = self.unique_columns(df)
        df.to_json(buffer, orient='records')
        buffer.seek(0)
        return buffer.getvalue()

    def row_to_json(self, data) -> str:
        return self._encoder.dumps(data)

    def row_to_string(self, data: dict) -> str:
        results = []
        for key, val in data.items():
            results.append(f'{key}: "{val!s}"')
        return ', '.join(results)

    def _load_excel(self, path: PurePath) -> list:
        """
        Load an Excel file using the Pandas library.

        Args:
            path (Path): The path to the Excel file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading Excel file: {path}")
            df = self.open_excel(path)
            # Check for row unicity:
            df = self.unique_columns(df)
            metadata = {
                "url": '',
                "index": '',
                "source": str(path.name),
                "filename": path.name,
                "question": '',
                "answer": '',
                "summary": '',
                "source_type": self._source_type,
                "type": self._type,
            }
            document_meta = {
                "columns": df.columns.tolist(),
                "rows": str(len(df)),
                "sheet": self.sheet_name,
            }
            documents = []
            # remove NaN values:
            df.fillna('', axis=1, inplace=True)
            for idx, row in df.iterrows():
                idk = ''
                rw = row.to_dict()
                for col in self._index_keys:
                    if col in row:
                        idk = row[col]
                        break
                _data = {
                    "index": idk,
                    "data": rw,
                    "document_meta": {
                        "row_index": idx,
                        **document_meta,
                        **rw
                    }
                }
                _meta = {**metadata, **_data}
                row_data = self.row_to_string(rw)
                doc = Document(
                    page_content=row_data,
                    metadata=_meta
                )
                documents.append(doc)
            return documents
        return []

    def load(self, max_tokens: int = 768) -> list:
        documents = []
        if self.path.is_file():
            # single File:
            documents = self._load_excel(self.path)
        elif self.path.is_dir():
            documents = []
            # iterate over the files in the directory
            for ext in self._extension:
                for item in self.path.glob(f'*{ext}'):
                    documents.extend(self._load_excel(item))
        elif isinstance(self.path, list):
            pass
        # Split Table Data:
        return self.split_documents(documents)

    def parse(self, source):
        raise NotImplementedError(
            "Parser method is not implemented for ExcelLoader."
        )



from abc import ABC
from typing import List
from langchain_core.prompts import ChatPromptTemplate


class AbstractLLM(ABC):
    """Abstract Language Model class.
    """

    model: str = "databricks/dolly-v2-3b"
    supported_models: List[str] = []
    embed_model: str = None
    max_tokens: int = 1024
    max_retries: int = 3

    @classmethod
    def get_supported_models(cls):
        return cls.supported_models

    def __init__(self, *args, **kwargs):
        self.model = kwargs.get("model", "databricks/dolly-v2-3b")
        self.task = kwargs.get("task", "text-generation")
        self.temperature: float = kwargs.get('temperature', 0.6)
        self.max_tokens: int = kwargs.get('max_tokens', 1024)
        self.top_k: float = kwargs.get('top_k', 10)
        self.top_p: float = kwargs.get('top_p', 0.90)
        self.args = {
            "top_p": self.top_p,
            "top_k": self.top_k,
        }
        self._llm = None
        self._embed = None

    def get_llm(self):
        return self._llm

    def get_embedding(self):
        return self._embed

    def __call__(self, text: str, **kwargs):
        return self._llm.invoke(text, **kwargs)

    def get_prompt(self, system: tuple, human: str) -> ChatPromptTemplate:
        """Get a prompt for the LLM."""
        return ChatPromptTemplate.from_messages(
            [("system", system), ("human", human)]
        )
